import logging
from abc import ABC
from pathlib import Path
from typing import Annotated, Dict, List, Literal, Optional, Set

from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    field_validator,
    model_validator,
)
from pydantic.types import NonNegativeFloat, PositiveInt
from strenum import StrEnum

from hydrolib.core.base.models import DiskOnlyFileModel
from hydrolib.core.dflowfm.common import LocationType
from hydrolib.core.dflowfm.common.models import Operand
from hydrolib.core.dflowfm.ini.io_models import Section
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    enum_value_parser,
    make_list,
    validate_required_fields,
)

logger = logging.getLogger(__name__)


class DataFileType(StrEnum):
    """
    Enum class containing the valid values for the dataFileType
    attribute in several subclasses of AbstractIniField.
    """

    arcinfo = "arcinfo"
    geotiff = "GeoTIFF"
    sample = "sample"
    onedfield = "1dField"
    polygon = "polygon"
    uniform = "uniform"
    netcdf = "netcdf"

    allowedvaluestext = "Possible values: arcinfo, GeoTIFF, sample, 1dField, polygon."


class InterpolationMethod(StrEnum):
    """
    Enum class containing the valid values for the interpolationMethod
    attribute in several subclasses of AbstractIniField.
    """

    constant = "constant"  # only with dataFileType=polygon .
    triangulation = "triangulation"  # Delaunay triangulation+linear interpolation.
    averaging = "averaging"  # grid cell averaging.
    linear_space_time = "linearSpaceTime"  # linear interpolation in space and time.

    allowedvaluestext = "Possible values: constant, triangulation, averaging."


class AveragingType(StrEnum):
    """
    Enum class containing the valid values for the averagingType
    attribute in several subclasses of AbstractIniField.
    """

    mean = "mean"  # simple average
    nearestnb = "nearestNb"  # nearest neighbour value
    max = "max"  # highest
    min = "min"  # lowest
    invdist = "invDist"  # inverse-weighted distance average
    minabs = "minAbs"  # smallest absolute value
    median = "median"

    allowedvaluestext = (
        "Possible values: mean, nearestNb, max, min, invDist, minAbs, median."
    )


class IniFieldGeneral(INIGeneral):
    """The initial field file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'iniField'. Do not edit this.",
            alias="fileType",
        )

    comments: Comments = Comments()
    _header: Literal["General"] = "General"
    fileversion: str = Field("2.00", alias="fileVersion")
    filetype: Literal["iniField"] = Field("iniField", alias="fileType")


class AbstractSpatialField(INIBasedModel, ABC):
    """
    Abstract base class for `[Initial]` and `[Parameter]` block data in
    inifield files.

    Defines all common fields. Used via subclasses InitialField and ParameterField.
    """

    class Comments(INIBasedModel.Comments):
        quantity: Optional[str] = Field(
            "Name of the quantity. See UM Table D.2.", alias="quantity"
        )
        datafile: Optional[str] = Field(
            "Name of file containing field data values.", alias="dataFile"
        )
        datafiletype: Optional[str] = Field("Type of dataFile.", alias="dataFileType")
        interpolationmethod: Optional[str] = Field(
            "Type of (spatial) interpolation.", alias="interpolationmethod"
        )
        operand: Optional[str] = Field(
            "How this data is combined with previous data for the same quantity (if any).",
            alias="operand",
        )
        averagingtype: Optional[str] = Field(
            "Type of averaging, if interpolationMethod=averaging .",
            alias="averagingtype",
        )
        averagingrelsize: Optional[str] = Field(
            "Relative search cell size for averaging.", alias="averagingrelsize"
        )
        averagingnummin: Optional[str] = Field(
            "Minimum number of points in averaging. Must be ≥ 1.",
            alias="averagingnummin",
        )
        averagingpercentile: Optional[str] = Field(
            "Percentile value for which data values to include in averaging. 0.0 means off.",
            alias="averagingpercentile",
        )
        extrapolationmethod: Optional[str] = Field(
            "Option for (spatial) extrapolation.", alias="extrapolationmethod"
        )
        locationtype: Optional[str] = Field(
            "Target location of interpolation.", alias="locationtype"
        )
        value: Optional[str] = Field(
            "Only for dataFileType=polygon. The constant value to be set inside for all model points inside the polygon."
        )

    comments: Comments = Comments()

    quantity: str = Field(alias="quantity")
    datafile: DiskOnlyFileModel = Field(alias="dataFile")

    datafiletype: DataFileType = Field(alias="dataFileType")
    interpolationmethod: Optional[InterpolationMethod] = Field(
        None, alias="interpolationMethod"
    )
    operand: Optional[Operand] = Field(Operand.override.value, alias="operand")
    averagingtype: Optional[AveragingType] = Field(None, alias="averagingType")
    averagingrelsize: Optional[NonNegativeFloat] = Field(None, alias="averagingRelSize")
    averagingnummin: Optional[PositiveInt] = Field(None, alias="averagingNumMin")
    averagingpercentile: Optional[NonNegativeFloat] = Field(
        None, alias="averagingPercentile"
    )
    extrapolationmethod: Optional[bool] = Field(False, alias="extrapolationMethod")
    locationtype: Optional[LocationType] = Field(
        LocationType.all.value, alias="locationType"
    )
    value: Optional[float] = Field(None, alias="value")

    model_config = ConfigDict(extra="allow")

    @classmethod
    def _process_section_values(cls, values):
        """Process Section objects and extract/convert values as needed.

        Args:
            values: The values to process, which may be a Section object or a dictionary.

        Returns:
            A dictionary containing the processed values.
        """
        # If values is a Section object, we need to handle it specially
        if isinstance(values, Section):
            # Extract the datafile value if present
            data_file = super()._extract_file_model_from_section(
                values, "datafile", DiskOnlyFileModel
            )

            # Convert Section to dictionary
            values_dict = super()._convert_section_to_dict(values)

            # If we found a datafile, add it to the dictionary
            if data_file is not None:
                values_dict["datafile"] = data_file

            return values_dict

        return values

    tracerfallvelocity: Optional[float] = Field(None, alias="tracerFallVelocity")
    tracerdecaytime: Optional[float] = Field(None, alias="tracerDecayTime")

    @model_validator(mode="before")
    @classmethod
    def validate_that_value_is_present_for_polygons(cls, values: Dict) -> Dict:
        """Validates that the value is provided when dealing with polygons."""
        # Process Section objects if needed
        values = cls._process_section_values(values)

        # Process dictionary-like objects
        data_file = values.get("datafile")
        if isinstance(data_file, (str, Path)):
            data_file = DiskOnlyFileModel(data_file)
            values["datafile"] = data_file

        validate_required_fields(
            values,
            "value",
            conditional_field_name="datafiletype",
            conditional_value=DataFileType.polygon,
        )

        value_field_value = values.get("value")
        datafiletype_field_value = values.get("datafiletype")
        if (
            value_field_value is not None
            and datafiletype_field_value is not None
            and datafiletype_field_value.lower() != DataFileType.polygon
        ):
            raise ValueError(
                f"When value={value_field_value} is given, dataFileType={DataFileType.polygon} is required."
            )

        return values

    @field_validator("locationtype", mode="before")
    @classmethod
    def validate_location_type(cls, v):
        return enum_value_parser(v, LocationType)

    @field_validator("averagingtype", mode="before")
    @classmethod
    def validate_average_type(cls, v):
        return enum_value_parser(v, AveragingType)

    @field_validator("datafiletype", mode="before")
    @classmethod
    def validate_data_file_type(cls, v):
        return enum_value_parser(v, DataFileType)

    @field_validator("operand", mode="before")
    @classmethod
    def validate_operand(cls, v):
        return enum_value_parser(v, Operand)

    @field_validator("interpolationmethod", mode="before")
    @classmethod
    def validate_interpolation_method(cls, v):
        return enum_value_parser(v, InterpolationMethod)

    @field_validator("datafile", mode="before")
    @classmethod
    def validate_datafile(cls, v):
        """Convert string values to DiskOnlyFileModel instances."""
        if isinstance(v, (str, Path)):
            return DiskOnlyFileModel(filepath=v)
        return v


class InitialField(AbstractSpatialField):
    """
    Initial condition field definition, represents an `[Initial]` block in
    an inifield file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile.initial[..]`

    All lowercased attributes match with the initial field input as described in
    [UM Sec.D.2](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.D.2).
    """

    _header: Literal["Initial"] = "Initial"


class ParameterField(AbstractSpatialField):
    """
    Parameter field definition, represents a `[Parameter]` block in
    an inifield file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile.parameter[..]`
    """

    _header: Literal["Parameter"] = "Parameter"


class IniFieldModel(INIModel):
    """
    The overall inifield model that contains the contents of one initial field and parameter file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile[..]`.

    Attributes:
        general (IniFieldGeneral): `[General]` block with file metadata.
        initial (List[InitialField]): List of `[Initial]` blocks with initial condition definitions.
        parameter (List[ParameterField]): List of `[Parameter]` blocks with spatial parameter definitions.
    """

    general: IniFieldGeneral = IniFieldGeneral()
    initial: Annotated[List[InitialField], BeforeValidator(make_list)] = Field(
        default_factory=list
    )
    parameter: Annotated[List[ParameterField], BeforeValidator(make_list)] = Field(
        default_factory=list
    )

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "fieldFile"



from .models import (
    ResearchFMModel,
    ResearchGeneral,
    ResearchGeometry,
    ResearchNumerics,
    ResearchOutput,
    ResearchPhysics,
    ResearchProcesses,
    ResearchRestart,
    ResearchSediment,
    ResearchSedtrails,
    ResearchTime,
    ResearchTrachytopes,
    ResearchWaves,
    ResearchWind,
)

__all__ = [
    "ResearchFMModel",
    "ResearchGeneral",
    "ResearchGeometry",
    "ResearchNumerics",
    "ResearchOutput",
    "ResearchPhysics",
    "ResearchProcesses",
    "ResearchRestart",
    "ResearchSediment",
    "ResearchTime",
    "ResearchTrachytopes",
    "ResearchWaves",
    "ResearchWind",
    "ResearchSedtrails",
]



from pathlib import Path
from typing import Any, Dict, List

from hydrolib.core.base.models import ModelSaveSettings


class CMPSerializer:
    @staticmethod
    def serialize(
        path: Path,
        data: Dict[str, List[Any]],
        save_settings: ModelSaveSettings,
    ) -> None:
        """
        Serialize components data to a file in .cmp format.

        Args:
            path (Path): The path to the destination .cmp file.
            data (Dict[str, Any]): A dictionary with keys "comments" and "components".
            - "comments" represents comments found at the start of the file.
            - "components" is a dictionary with keys "harmonics" and "astronomics".
                - "harmonics" is a list of dictionaries with the values as "period", "amplitude" and "phase".
                    - "period" is a float as a string.
                    - "amplitude" is a float as a string.
                    - "phase" is a float as a string.
                - "astronomics" is a list of dictionaries with the values as "name", "amplitude" and "phase".
                    - "name" is an AstronomicName as a string.
                    - "amplitude" is a float as a string.
                    - "phase" is a float as a string.

            save_settings (ModelSaveSettings): The save settings to be used.
        """
        path.parent.mkdir(parents=True, exist_ok=True)

        comment_lines = CMPSerializer._serialize_comment_lines(data)
        component_lines = CMPSerializer._serialize_components_lines(data["component"])

        file_content = CMPSerializer._serialize_file_content(
            component_lines, comment_lines
        )
        with path.open("w", encoding="utf8") as file:
            file.write(file_content)

    @staticmethod
    def _serialize_comment_lines(data: Dict[str, List[Any]]) -> List[str]:
        comment_lines = []
        for comment in data["comments"]:
            comment_lines.append(f"{'#' if comment != '' else ''}{comment}")
        return comment_lines

    @staticmethod
    def _serialize_components_lines(data: Dict[str, Any]) -> List[str]:
        components_lines = []
        if "harmonics" in data:
            for harmonic in data["harmonics"]:
                components_lines.append(
                    f"{harmonic['period']} {harmonic['amplitude']} {harmonic['phase']}"
                )
        if "astronomics" in data:
            for astronomic in data["astronomics"]:
                components_lines.append(
                    f"{astronomic['name']} {astronomic['amplitude']} {astronomic['phase']}"
                )
        return components_lines

    @staticmethod
    def _serialize_file_content(
        components_lines: List[str], comment_lines: List[str]
    ) -> str:
        return "\n".join(comment_lines + components_lines)



"""External forcing converter."""

import os
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from hydrolib.core.base.file_manager import PathOrStr, resolve_relative_to_root
from hydrolib.core.base.models import DiskOnlyFileModel
from hydrolib.core.dflowfm.bc.models import (
    T3D,
    Astronomic,
    ForcingBase,
    ForcingModel,
    Harmonic,
    QuantityUnitPair,
    TimeSeries,
)
from hydrolib.core.dflowfm.cmp.models import AstronomicRecord, CMPModel, HarmonicRecord
from hydrolib.core.dflowfm.ext.models import (
    SOURCE_SINKS_QUANTITIES_VALID_PREFIXES,
    Boundary,
    BoundaryError,
    InitialFieldError,
    Meteo,
    MeteoError,
    SourceSink,
    SourceSinkError,
)
from hydrolib.core.dflowfm.extold.models import (
    ExtOldBoundaryQuantity,
    ExtOldForcing,
    ExtOldInitialConditionQuantity,
    ExtOldMeteoQuantity,
    ExtOldParametersQuantity,
    ExtOldSourcesSinks,
)
from hydrolib.core.dflowfm.inifield.models import InitialField, ParameterField
from hydrolib.core.dflowfm.polyfile.models import PolyFile
from hydrolib.core.dflowfm.t3d.models import T3DModel
from hydrolib.core.dflowfm.tim.models import TimModel
from hydrolib.core.dflowfm.tim.parser import TimParser
from hydrolib.tools.extforce_convert.utils import (
    convert_interpolation_data,
    create_initial_cond_and_parameter_input_dict,
    find_temperature_salinity_in_quantities,
    oldfiletype_to_forcing_file_type,
)


class BaseConverter(ABC):
    """Abstract base class for converting old external forcings blocks to new blocks.

    Subclasses must implement the `convert` method, specific for the
    type of model data in the various old external forcing blocks.

    Class ConverterFactory uses these subclasses to create the correct
    converter, depending on the quantity of the forcing block.
    """

    def __init__(self):
        """Initializes the BaseConverter object."""
        self._root_dir = None
        self._legacy_files = []

    @property
    def root_dir(self) -> Path:
        """Get the root directory of the external forcing files."""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, value: Union[Path, str]):
        if isinstance(value, str):
            value = Path(value)
        self._root_dir = value

    @property
    def legacy_files(self) -> List[Path]:
        return self._legacy_files

    @legacy_files.setter
    def legacy_files(self, value: Union[PathOrStr, List[PathOrStr]]):
        """Set the legacy files to be cleaned up after conversion."""
        if isinstance(value, list):
            self._legacy_files += [Path(file) for file in value]
        else:
            self._legacy_files += [Path(value)]

    @abstractmethod
    def convert(self, forcing: ExtOldForcing) -> Any:
        """Converts the data from the old external forcings format to the proper/new model input block.

        Args:
            forcing (ExtOldForcing): The data read from an old format
                external forcings file.

        Returns:
            Any: The converted data in the new format. Should be
                included in some FileModel object by the caller.
        """
        raise NotImplementedError("Subclasses must implement convert method")


class MeteoConverter(BaseConverter):
    """Meteo quantities Converter."""

    def __init__(self):
        """Meteo converter constructor."""
        super().__init__()

    def convert(self, forcing: ExtOldForcing) -> Meteo:
        """Meteo converter.

        Convert an old external forcing block with meteo data to a Meteo
        forcing block suitable for inclusion in a new external forcings file.

        This function takes a forcing block from an old external forcings
        file, represented by an instance of ExtOldForcing, and converts it
        into a Meteo object. The Meteo object is suitable for use in new
        external forcings files, adhering to the updated format and
        specifications.

        Args:
            forcing (ExtOldForcing): The contents of a single forcing block
            in an old external forcings file. This object contains all the
            necessary information, such as quantity, values, and timestamps,
            required for the conversion process.

        Returns:
            Meteo: A Meteo object that represents the converted forcing
            block, ready to be included in a new external forcings file. The
            Meteo object conforms to the new format specifications, ensuring
            compatibility with updated systems and models.

        Raises:
            ValueError: If the forcing block contains a quantity that is not
            supported by the converter, a ValueError is raised. This ensures
            that only compatible forcing blocks are processed, maintaining
            data integrity and preventing errors in the conversion process.
        """
        meteo_data = {
            "quantity": forcing.quantity,
            "forcingfile": forcing.filename,
            "forcingfiletype": oldfiletype_to_forcing_file_type(forcing.filetype),
            "forcingVariableName": forcing.varname,
        }
        if forcing.sourcemask != DiskOnlyFileModel(None):
            raise ValueError(
                f"Attribute 'SOURCEMASK' is no longer supported, cannot "
                f"convert this input. Encountered for QUANTITY="
                f"{forcing.quantity} and FILENAME={forcing.filename}."
            )
        meteo_data = convert_interpolation_data(forcing, meteo_data)
        meteo_data["extrapolationAllowed"] = bool(forcing.extrapolation_method)
        meteo_data["extrapolationSearchRadius"] = forcing.maxsearchradius
        meteo_data["operand"] = forcing.operand
        try:
            meteo_block = Meteo(**meteo_data)
        except Exception as e:
            raise MeteoError(
                f"Failed to create the Meteo object. for the following Errors: {e}"
            )

        return meteo_block


class BoundaryConditionConverter(BaseConverter):
    """Boundary condition converter."""

    def __init__(self):
        """Boundary condition converter constructor."""
        super().__init__()

    @staticmethod
    def merge_tim_files(tim_files: List[Path], quantity: str) -> TimModel:
        """Parse the boundary condition related time series from the tim files.

        The function will merge all the tim files into one tim model and assign the quantity names to the tim model.

        Args:
            tim_files (List[Path]):
                List of TIM models paths.
            quantity (str):
                name of the quantity that the tim files represent.
        Returns:
            TimModel: A TimModel object containing the time series data from all given TIM files.
        """
        time_files_exist = all([tim_file.exists() for tim_file in tim_files])
        if not time_files_exist:
            raise FileNotFoundError(
                f"TIM files '{tim_files}' not found for QUANTITY={quantity}"
            )

        tim_models = [
            TimModel(file, quantities_names=[file.stem]) for file in tim_files
        ]
        # merge all the tim files into one tim model
        for tim_model in tim_models[1:]:
            data = tim_model.as_dict()
            if len(data.keys()) != 1:
                raise ValueError(
                    f"Number of columns in the TIM file '{tim_model.filepath}' should be 1 column. in addition to the "
                    "time column."
                )
            tim_models[0].add_column(
                list(data.values())[0], column_name=list(data.keys())[0]
            )
        return tim_models[0]

    def convert_tim_to_bc(
        self,
        tim_files: List[PathOrStr],
        time_unit: str,
        time_interpolation: str = "linear",
        quantity: str = None,
        label: str = None,
    ) -> List[TimeSeries]:
        """Convert a TimModel into a ForcingModel.

        wrapper on top of the `TimToForcingConverter.convert` method. to customize it for the source and sink

        Args:
            tim_files (List[Union[Path, str]]):
                paths to the tim files to be converted.
            time_unit (str):
                Formatted string containing the units of time, including absolute datetime reference information
                (according to UDunits). For example, "minutes since 1992-10-8 15:15:42.5 -6:00".
            time_interpolation (str, default is linear):
                The interpolation method to be used for the time series data.
            quantity (str, default is None):
                name of the quantity that the tim files represent.
            label (str, default is None):
                the label from the pli file to be used to name the time series sections in the .bc model.

        Returns:
            ForcingModel: The converted ForcingModel.

        Raises:
            ValueError: If `units` and `user_defined_names` are not provided.
            ValueError: If the lengths of `units`, `user_defined_names`, and the columns in the first row of the TimModel
        """
        tim_model = self.merge_tim_files(tim_files, quantity)

        # switch the quantity names from the Tim model (loction names) to quantity names.
        user_defined_names = BoundaryConditionConverter._get_file_labels(
            label, tim_files
        )
        tim_model.quantities_names = [quantity] * len(tim_model.get_units())

        units = tim_model.get_units()
        time_series_list = TimToForcingConverter.convert(
            tim_model, time_unit, time_interpolation, units, user_defined_names
        )
        return time_series_list

    def locate_files(self, location_file: Path):
        """Locate the tim, t3d, and cmp files related to the location file.

        Args:
            location_file(Path):
                the pli file that contains the location of the boundary condition.

        Returns:
            tim_files (List[Path]):
                list of all the tim files related to the location file.
            t3d_files (List[Path]):
                list of all the t3d files related to the location file.
            cmp_files (List[Path]):
                list of all the cmp files related to the location file.
        """
        forcings_local_dir = resolve_relative_to_root(location_file, self.root_dir)
        FILE_NUMBERING_PATTERN = "[0-9][0-9][0-9][0-9]*"
        stem_pattern = f"{location_file.stem}_{FILE_NUMBERING_PATTERN}"
        tim_files = list(forcings_local_dir.parent.glob(f"{stem_pattern}.tim"))
        t3d_files = list(forcings_local_dir.parent.glob(f"{stem_pattern}.t3d"))
        cmp_files = list(forcings_local_dir.parent.glob(f"{stem_pattern}.cmp"))
        return tim_files, t3d_files, cmp_files

    def convert(
        self, forcing: ExtOldForcing, time_unit: Optional[str] = None
    ) -> Boundary:
        """Boundary condition converter.

        Convert an old external forcing block to a boundary forcing block
        suitable for inclusion in a new external forcings file.

        This function takes a forcing block from an old external forcings
        file, represented by an instance of ExtOldForcing, and converts it
        into a boundary object. The Boundary object is suitable for use in new
        external forcings files, adhering to the updated format and
        specifications.

        Args:
            forcing (ExtOldForcing): The contents of a single forcing block
                in an old external forcings file. This object contains all the
                necessary information, such as quantity, values, and timestamps,
                required for the conversion process.
            time_unit:
                The start date of the time series data.

        Returns:
            Boundary: A Boundary object that represents the converted forcing
            block, ready to be included in a new external forcings file. The
            Boundary object conforms to the new format specifications, ensuring
            compatibility with updated systems and models.

        Raises:
            ValueError: If the forcing block contains a quantity that is not
            supported by the converter, a ValueError is raised. This ensures
            that only compatible forcing blocks are processed, maintaining
            data integrity and preventing errors in the conversion process.

        Notes:
            - The `root_dir` property must be set before calling this method.
            - Since the `start_time` argument must be provided from the mdu file to convert the time series data,
            boundary Condition can be only converted by reading the mdu file and the external forcing file is not
            enough.
            - The new labels for all quantities in the .bc file will be taken from the pli file and the number at the
            end of the label is taken from the file name of the tim, t3d, or cmp files.
        """
        quantity = forcing.quantity
        location_file = forcing.filename.filepath
        poly_line = forcing.filename
        if not isinstance(poly_line, PolyFile):
            poly_line = PolyFile(location_file)

        label = poly_line.objects[0].metadata.name
        if self.root_dir is None:
            raise ValueError(
                "The 'root_dir' property must be set before calling this method."
            )

        tim_files, t3d_files, cmp_files = self.locate_files(location_file)
        forcings_list = []

        if len(tim_files) > 0:
            time_series_list = self.convert_tim_to_bc(
                tim_files, time_unit, quantity=quantity, label=label
            )
            forcings_list.extend(time_series_list)
            self.legacy_files = tim_files

        # check t3d files
        if len(t3d_files) > 0:
            t3d_forcing_list = self._convert_t3d_files(t3d_files, quantity, label)
            forcings_list.extend(t3d_forcing_list)
            self.legacy_files = t3d_files

        # check cmp files
        if len(cmp_files) > 0:
            forcing_list = self._convert_cmp_files(cmp_files, quantity, label)
            forcings_list.extend(forcing_list)
            self.legacy_files = cmp_files

        forcing_model = ForcingModel(forcing=forcings_list)

        # set the bc file names to the same names as the tim files.
        forcing_model.filepath = location_file.with_suffix(".bc")

        data = {
            "quantity": forcing.quantity,
            "locationfile": location_file,
            "forcingfile": forcing_model,
        }

        try:
            new_block = Boundary(**data)
        except Exception as e:
            raise BoundaryError(
                f"Failed to create the Boundary object. for the following Errors: {e}"
            )

        return new_block

    @staticmethod
    def _convert_t3d_files(
        t3d_files: List[Path], quantity: str, label: str
    ) -> List[T3D]:
        """Convert T3D files to T3D forcing objects.

        Args:
            t3d_files (List[Path]):
                t3d files to be converted.
            quantity (str):
                quantity name that the t3d files represent.
            label (str):
                label from the pli file to be used to name the time series sections in the .bc model.

        Returns:
            List[T3D]:
                A list of T3D objects representing the converted T3D files.
        """
        t3d_models = [T3DModel(path) for path in t3d_files]
        # this line assumed that the two t3d files will have the same number of layers and same number of quantities
        quantities_names = [quantity] * t3d_models[0].size[1]
        user_defined_names = BoundaryConditionConverter._get_file_labels(
            label, t3d_files
        )
        t3d_forcing_list = T3DToForcingConverter.convert(
            t3d_models, quantities_names, user_defined_names
        )
        return t3d_forcing_list

    @staticmethod
    def _convert_cmp_files(
        cmp_files: List[Path], quantity: str, label: str
    ) -> List[ForcingBase]:
        """Convert CMP files to ForcingModel.

        Args:
            cmp_files (List[Path]):
                List of CMP files to be converted.
            quantity (str):
                quantity name that the cmp files represent.
            label (str):
                label from the pli file names to be used to name the time series sections in the .bc model.

        Returns:
            List[ForcingBase]:
                The converted ForcingBase object.
        """
        cmp_models = [CMPModel(path) for path in cmp_files]
        user_defined_names = BoundaryConditionConverter._get_file_labels(
            label, cmp_files
        )

        for cmp_model in cmp_models:
            cmp_model.quantities_name = [quantity]
        forcing_list = CMPToForcingConverter.convert(cmp_models, user_defined_names)
        return forcing_list

    @staticmethod
    def _get_file_labels(label: str, files: List[Path]) -> List[str]:
        """
        Get the labels of the files based on their filenames and a provided label.

        Args:
            label (str):
                A string label to prefix the generated file labels.
            files (List[Path]):
                A list of file paths. Each file's name is expected to end with '_<number>',
                where <number> is an integer used to generate the labels.

        Returns:
            List[str]:
                A list of strings representing the labels for the files. Each label is
                generated by appending '_<number>' (zero-padded to 4 digits) to the provided label.

        Assumptions:
            - Filenames must end with '_<number>', where <number> is an integer.
            - The method extracts this integer from the filename to generate the labels.
        """
        try:
            file_int = [file.stem.split("_")[-1] for file in files]
            file_int_id = [int(i) for i in file_int]
        except ValueError:
            raise ValueError(
                f"Cannot get the file number from the file name. file name should be <NAME>_<INT-NUMBER> "
                f"Please check the file names: {files}"
            )
        user_defined_names = [f"{label}_{str(i).zfill(4)}" for i in file_int_id]
        return user_defined_names


class InitialConditionConverter(BaseConverter):
    """Initial condition converter."""

    def __init__(self):
        """Initial condition converter constructor."""
        super().__init__()

    def convert(self, forcing: ExtOldForcing, new_forcing_path: Path) -> InitialField:
        """Convert the Initial condition quantities.

        Convert an old external forcing block with Initial condition data to a IinitialField
        forcing block suitable for inclusion in a new inifieldfile file.


        This function takes a forcing block from an old external forcings
        file, represented by an instance of ExtOldForcing, and converts it
        into a InitialField object. The InitialField object is suitable for use in new
        iniFieldFile, adhering to the updated format and specifications.

        Args:
            forcing (ExtOldForcing):
                The contents of a single forcing block
                in an old external forcings file. This object contains all the
                necessary information, such as quantity, values, and timestamps,
                required for the conversion process.
            new_forcing_path (Path):
                The updated path to the forcing data file.

        Returns:
            Initial condition field definition, represents an `[Initial]` block in an inifield file.

        Raises:
            ValueError: If the forcing block contains a quantity that is not
            supported by the converter, a ValueError is raised. This ensures
            that only compatible forcing blocks are processed, maintaining
            data integrity and preventing errors in the conversion process.

        References:
            [Sec.D](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.D)
        """
        data = create_initial_cond_and_parameter_input_dict(forcing, new_forcing_path)
        try:
            new_block = InitialField(**data)
        except Exception as e:
            raise InitialFieldError(
                f"Failed to create the InitialField object. for the following Errors: {e}"
            )

        return new_block


class ParametersConverter(BaseConverter):
    """Parameter converter."""

    def __init__(self):
        """Parameter converter constructor."""
        super().__init__()

    def convert(self, forcing: ExtOldForcing, new_forcing_path: Path) -> ParameterField:
        """Parameter converter.

        Convert an old external forcing block to a parameter forcing block
        suitable for inclusion in an initial field and parameter file.

        This function takes a forcing block from an old external forcings
        file, represented by an instance of ExtOldForcing, and converts it
        into a ParameterField object. The ParameterField object is suitable for use in
        an IniFieldModel, representing an initial field and parameter file, adhering
        to the updated format and specifications.

        Args:
            forcing (ExtOldForcing):
                The contents of a single forcing block
                in an old external forcings file. This object contains all the
                necessary information, such as quantity, values, and timestamps,
                required for the conversion process.
            new_forcing_path (Path):
                The updated path to the forcing data file.

        Returns:
            ParameterField:
                A ParameterField object that represents the converted forcing
                block, ready to be included in an initial field and parameter file. The
                ParameterField object conforms to the new format specifications, ensuring
                compatibility with updated systems and models.

        Raises:
            ValueError: If the forcing block contains a quantity that is not
            supported by the converter, a ValueError is raised. This ensures
            that only compatible forcing blocks are processed, maintaining
            data integrity and preventing errors in the conversion process.
        """
        data = create_initial_cond_and_parameter_input_dict(forcing, new_forcing_path)
        new_block = ParameterField(**data)

        return new_block


class SourceSinkConverter(BaseConverter):
    """Source and sink converter."""

    def __init__(self):
        """Source and sink converter Constructor."""
        super().__init__()

    @staticmethod
    def merge_mdu_and_ext_file_quantities(
        mdu_quantities: Dict[str, bool], temp_salinity_from_ext: Dict[str, int]
    ) -> List[str]:
        """Merge the temperature and salinity from the mdu file with the temperature and salinity from the external file.

        Args:
            mdu_quantities (Dict[str, bool]): A dictionary containing the temperature and salinity details from the
                mdu file, with bool values indecating if the temperature/salinity is activated in the mdu file.
            temp_salinity_from_ext (Dict[str,int]): A dictionary containing the temperature and salinity details from
                the external file.

        Returns:
            List[str]: A list of quantities that will be used in the tim file.
        """
        if mdu_quantities:
            mdu_file_quantity_list = [key for key, val in mdu_quantities.items() if val]
            temp_salinity_from_mdu = find_temperature_salinity_in_quantities(
                mdu_file_quantity_list
            )
            final_temp_salinity = temp_salinity_from_ext | temp_salinity_from_mdu
            # the kwargs will be provided only from the source and sink converter
            # Ensure 'temperature' comes before 'salinity'
            keys = list(final_temp_salinity.keys())
            if (
                "sourcesink_temperaturedelta" in keys
                and "sourcesink_salinitydelta" in keys
            ):
                keys.remove("sourcesink_salinitydelta")
                keys.insert(
                    keys.index("sourcesink_temperaturedelta"),
                    "sourcesink_salinitydelta",
                )
        else:
            keys = list(temp_salinity_from_ext.keys())

        return keys

    def parse_tim_model(
        self, tim_file: Path, ext_file_quantity_list: List[str], **mdu_quantities
    ) -> TimModel:
        """Parse the source and sinks related time series from the tim file.

        - Parse the TIM file and extract the time series data for each column.
        - assign the time series data to the corresponding quantity name.

        The order of the quantities in the tim file should be as follows:
        - time
        - sourcesink_discharge
        - sourcesink_salinitydelta (optional)
        - sourcesink_temperatureDelta (optional)
        - tracer<anyname>delta (optional)
        - any other quantities from the external forcings file.

        Args:
            tim_file (Path): The path to the TIM file.
            ext_file_quantity_list (List[str]): A list of other quantities that are present in the external forcings file.
            **mdu_quantities: keyword argumens that will be provided if you want to provide the temperature and salinity
                details from the mdu file, the dictionary will have two keys `temperature`, `salinity` and the values are
                only bool. (i.e. {"temperature", False, "salinity": True})

        Returns:
            TimeModel: The same `TimModel after assigning the quantity names,  the time series data form each column in
            the tim_file.
            the keys of the dictionary will be the quantity names, and the values will be the time series data.

        Raises:
            ValueError: If the number of columns in the TIM file does not match the number of quantities in the external
            forcings file that has one of the following prefixes `initialtracer`,`tracerbnd`,
            `sedfracbnd`,`initialsedfrac`, plus the discharge, temperature, and salinity.

        Notes:
            - The function will combine the temperature and salinity from the MDU file (value is 1) file with the
                quantities mentioned in the external forcing file, and will get the list of quantities that are in the tim file.
            - The function will return a dictionary with the quantities as keys and the time series data as values.

        Examples:
        if the tim file contains 5 columns (the first column is the time):
            ```
            0.0 1.0 2.0 3.0 4.0
            1.0 1.0 2.0 3.0 4.0
            2.0 1.0 2.0 3.0 4.0
            3.0 1.0 2.0 3.0 4.0
            4.0 1.0 2.0 3.0 4.0
            ```
        and the external file contains the following quantities:
            >>> ext_file_quantity_list = ["discharge", "temperature", "salinity", "initialtracerAnyname",
            ... "anyother-quantities"]

        - The function will filter the external forcing quantities that have one of the following prefixes
        `initialtracer`,`tracerbnd`, `sedfracbnd`,`initialsedfrac`, plus the discharge, temperature, and salinity.
        - If the mdu_quantities are provided, the function will merge the temperature and salinity from the mdu file
        with the filtered quantities mentioned in the external forcing file.
        - The merged list of quantities from both the ext and mdu files will then be compared with the number of
        columns in the TIM file, if they don't match a `Value Error` will be raised.
        - Here the filtered quantities are ["discharge", "temperature", "salinity", "initialtracerAnyname"] and the
        tim file contains 4 columns (excluding the time column).
            ```python
            >>> from pathlib import Path
            >>> from hydrolib.tools.extforce_convert.converters import SourceSinkConverter
            >>> tim_file = Path("tests/data/input/source-sink/leftsor.tim")
            >>> converter = SourceSinkConverter()
            >>> tim_model = converter.parse_tim_model(tim_file, ext_file_quantity_list)
            >>> print(tim_model.quantities_names)
            ['sourcesink_discharge', 'sourcesink_salinitydelta', 'sourcesink_temperaturedelta', 'initialtracerAnyname']
            >>> print(tim_model.as_dict()) # doctest: +SKIP
            {
                "discharge": [1.0, 1.0, 1.0, 1.0, 1.0],
                "sourcesink_salinitydelta": [2.0, 2.0, 2.0, 2.0, 2.0],
                "sourcesink_temperatureDelta": [3.0, 3.0, 3.0, 3.0, 3.0],
                "initialtracerAnyname": [4.0, 4.0, 4.0, 4.0, 4.0],
            }

            ```

        mdu file:
            ```ini
            [physics]
            ...
            Salinity             = 1        # Include salinity, (0=no, 1=yes)
            ...
            Temperature          = 1        # Include temperature, (0=no, 1=only transport, 3=excess model of D3D,5=heat flux model (5) of D3D)
            ```

        external forcings file:
            ```
            QUANTITY=initialtemperature
            FILENAME=right.pol
            ...

            QUANTITY=initialsalinity
            FILENAME=right.pol
            ...
            ```
        """
        time_file = TimParser.parse(tim_file)
        tim_model = TimModel(**time_file)
        time_series = tim_model.as_dict()
        # get the required quantities from the external file
        required_quantities_from_ext = [
            key
            for key in ext_file_quantity_list
            if key.startswith(SOURCE_SINKS_QUANTITIES_VALID_PREFIXES)
        ]
        # Remove duplicate quantities that might be present in the list due to quantities that share names,
        # therefore occurring multiple times in the external forcing file.
        # TimeSeries columns are expected to be linked to unique quantity names.
        required_quantities_from_ext = list(set(required_quantities_from_ext))

        # check if the temperature and salinity are present in the external file
        temp_salinity_from_ext = find_temperature_salinity_in_quantities(
            ext_file_quantity_list
        )

        final_temp_salinity = self.merge_mdu_and_ext_file_quantities(
            mdu_quantities, temp_salinity_from_ext
        )
        final_quantities_list = (
            ["sourcesink_discharge"]
            + final_temp_salinity
            + required_quantities_from_ext
        )

        if len(time_series) != len(final_quantities_list):
            raise ValueError(
                f"Number of columns in the TIM file '{tim_file}: {len(time_series)}' does not match the number of "
                f"quantities in the external forcing file: {final_quantities_list}."
            )
        # assign the quantity names to the tim model
        tim_model.quantities_names = final_quantities_list
        return tim_model

    @staticmethod
    def convert_tim_to_bc(
        tim_model: TimModel,
        time_unit: str,
        user_defined_names: List[str] = None,
    ) -> ForcingModel:
        """Convert a TimModel into a ForcingModel.

            wrapper in top of the `TimToForcingConverter.convert` method. to customize it for the source and sink

        Args:
            tim_model (TimModel):
                The input TimModel to be converted.
            time_unit (str):
                Formatted string containing the units of time, including absolute datetime reference information
                (according to UDunits). For example, "minutes since 1992-10-8 15:15:42.5 -6:00".
            user_defined_names (List[str], optional):
                A list of user-defined names for the forcing blocks.

        Returns:
            ForcingModel: The converted ForcingModel.

        Raises:
            ValueError: If `units` and `user_defined_names` are not provided.
            ValueError: If the lengths of `units`, `user_defined_names`, and the columns in the first row of the TimModel
        """
        units = tim_model.get_units()
        time_series_list = TimToForcingConverter.convert(
            tim_model, time_unit, units=units, user_defined_names=user_defined_names
        )
        forcing_model = ForcingModel(forcing=time_series_list)
        return forcing_model

    @staticmethod
    def separate_forcing_model(forcing_model: ForcingModel) -> Dict[str, ForcingModel]:
        """Separate the forcing model into a list of forcing models.

        each forcing model will contain only one forcing quantity.
        """
        forcing_list = [deepcopy(forcing) for forcing in forcing_model.forcing]

        forcing_model_list = []
        forcings = {}
        for forcing in forcing_list:
            model = deepcopy(forcing_model)
            model.forcing = [forcing]
            forcing_model_list.append(model)
            name = forcing.quantityunitpair[1].quantity
            # remove the prefix 'sourcesink_' from the name as the extforce file will not have this prefix.
            forcings[name.removeprefix("sourcesink_")] = model

        return forcings

    def convert(
        self,
        forcing: ExtOldForcing,
        ext_file_quantity_list: List[str] = None,
        start_time: str = None,
        **temp_salinity_mdu,
    ) -> SourceSink:
        """Source and sink converter.

        Convert an old external forcing block with Sources and sinks to a SourceSink
        forcing block suitable for inclusion in a new external forcings file.

        Args:
            forcing (ExtOldForcing): The contents of a single forcing block in an old external forcings file. This
                object contains all the necessary information, such as quantity, values, and timestamps, required for the
                conversion process.
            ext_file_quantity_list (List[str], default is None): A list of other quantities that are present in the
                external forcings file.
            start_time (str, default is None):
                The start date of the time series data.
            **temp_salinity_mdu:
                keyword arguments that will be provided if you want to provide the temperature and salinity details from
                the mdu file, the dictionary will have two keys `temperature`, `salinity` and the values are only bool.
                ```python
                {'salinity': True, 'temperature': True}
                ```

        Returns:
            SourceSink: A SourceSink object that represents the converted forcing
            block, ready to be included in a new external forcings file. The
            SourceSink object conforms to the new format specifications, ensuring
            compatibility with updated systems and models.

        Raises:
            ValueError: If the forcing block contains a quantity that is not
            supported by the converter, a ValueError is raised. This ensures
            that only compatible forcing blocks are processed, maintaining
            data integrity and preventing errors in the conversion process.

        Notes:
            - Since the `start_time` argument must be provided from the mdu file to convert the time series data,
            SourceSink can be only converted by reading the mdu file and the external forcing file is not
            enough.

        References:
            - `Sources and Sinks <https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#C10>`_
            - `Polyline <https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#C2>`
            - `TIM file format <https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#C4>`_
            - `Sources and Sinks <https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#5.4.10>`_
            - `Source and sink definitions <https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#C5.2.4>`_

        """
        location_file = forcing.filename.filepath
        polyline = forcing.filename
        location_name = location_file.stem
        z_source, z_sink = polyline.get_z_sources_sinks()

        # check the tim file
        tim_file = resolve_relative_to_root(
            polyline.filepath, self.root_dir
        ).with_suffix(".tim")
        if not tim_file.exists():
            raise ValueError(
                f"TIM file '{tim_file}' not found for QUANTITY={forcing.quantity}"
            )

        self.legacy_files = tim_file

        tim_model = self.parse_tim_model(
            tim_file, ext_file_quantity_list, **temp_salinity_mdu
        )
        labels = [f"{location_name}"] * len(tim_model.quantities_names)

        forcing_model = self.convert_tim_to_bc(
            tim_model, start_time, user_defined_names=labels
        )
        # set the bc file names to the same names as the tim files.
        forcing_model.filepath = Path(
            os.path.relpath(tim_file, self.root_dir.resolve())
        ).with_suffix(".bc")

        data = {
            "id": location_name,
            "name": forcing.quantity,
            "numcoordinates": len(polyline.x),
            "xcoordinates": polyline.x,
            "ycoordinates": polyline.y,
        }
        if forcing.area is not None:
            data["area"] = forcing.area
        forcings = self.separate_forcing_model(forcing_model)

        # the same forcing model will be used for all the forcings to be able to save all the forcings (sourcesinks)
        # in the same file.
        for name, _ in forcings.items():
            forcings[name] = forcing_model

        data = data | forcings

        if None not in z_source:
            # if the z_source and z_sink are not None, then add them to the data
            z_source_sink_data = {
                "zsource": z_source,
                "zsink": z_sink,
            }
            data = data | z_source_sink_data

        try:
            new_block = SourceSink(**data)
        except Exception as e:  # pragma: no cover
            raise SourceSinkError(
                f"Failed to create the SourceSink object. for the following Errors: {e}"
            )

        return new_block


class ConverterFactory:
    """A factory class for creating converters based on the given quantity."""

    @staticmethod
    def create_converter(quantity) -> BaseConverter:
        """
        Create converter based on the given quantity.

        Args:
            quantity: The quantity for which the converter needs to be created.

        Returns:
            BaseConverter: An instance of a specific BaseConverter subclass
                for the given quantity.

        Raises:
            ValueError: If no converter is available for the given quantity.
        """
        if ConverterFactory.contains(ExtOldMeteoQuantity, quantity):
            return MeteoConverter()
        elif ConverterFactory.contains(ExtOldInitialConditionQuantity, quantity):
            return InitialConditionConverter()
        elif ConverterFactory.contains(ExtOldBoundaryQuantity, quantity):
            return BoundaryConditionConverter()
        elif ConverterFactory.contains(ExtOldParametersQuantity, quantity):
            return ParametersConverter()
        elif ConverterFactory.contains(ExtOldSourcesSinks, quantity):
            return SourceSinkConverter()
        else:
            raise ValueError(f"No converter available for QUANTITY={quantity}.")

    @staticmethod
    def contains(quantity_class, quantity) -> bool:
        """Check if the given quantity is in the specified class."""
        try:
            quantity_class(quantity)
        except ValueError:
            return False

        return True


class CMPToForcingConverter:
    """A class to convert CmpModel data into ForcingModel data for boundary condition definitions."""

    @staticmethod
    def convert(
        cmp_models: List[CMPModel], user_defined_names: List[str] = None
    ) -> List[ForcingBase]:
        """
        Convert a CmpModel into a ForcingModel.

        Args:
            cmp_models (List[CmpModel]):
                The input CmpModel to be converted.
            user_defined_names (List[str]):
                user defined names for the quantities.Default is None.

        Returns:
            ForcingModel: The converted ForcingModel.

        Raises:
            ValueError: If the lengths of the columns in the first row of the CmpModel do not match the number of
                quantities in the CmpModel.

        Examples:
            Convert a CmpModel into a ForcingModel.
                ```python
                >>> harmonic_data = {
                ...     "comments": ["# Example comment"],
                ...     "component": {
                ...         "harmonics": [{"period": 0.0, "amplitude": 1.0, "phase": 2.0}],
                ...     },
                ...     "quantities_name": ["discharge"],
                ... }
                >>> astronomic_data = {
                ...     "comments": ["# Example comment"],
                ...     "component": {
                ...         "astronomics": [{"name": "4MS10", "amplitude": 1.0, "phase": 2.0}],
                ...     },
                ...     "quantities_name": ["waterlevel"],
                ... }
                >>> cmp_models = [CMPModel(**harmonic_data), CMPModel(**astronomic_data)]
                >>> forcing_model = CMPToForcingConverter.convert(cmp_models,["L1_0001", "L1_0002"])
                >>> forcing_model[0].datablock
                [[0.0, 1.0, 2.0]]
                >>> forcing_model[1].datablock
                [['4MS10', 1.0, 2.0]]

                ```
        """
        forcing_list = []

        for label, cmp_model in zip(user_defined_names, cmp_models):
            if cmp_model.component.harmonics:
                harmonic_model = CMPToForcingConverter.convert_harmonic(
                    label,
                    cmp_model.component.harmonics,
                    cmp_model.quantities_name[0],
                    unit=cmp_model.get_units()[0],
                )
                forcing_list.append(harmonic_model)

            if cmp_model.component.astronomics:
                astronomic_model = CMPToForcingConverter.convert_astronomic(
                    label,
                    cmp_model.component.astronomics,
                    cmp_model.quantities_name[0],
                    unit=cmp_model.get_units()[0],
                )
                forcing_list.append(astronomic_model)

        return forcing_list

    @staticmethod
    def convert_harmonic(
        user_defined_name,
        harmonics: List[HarmonicRecord],
        quantity_name: str,
        unit: str,
    ) -> Harmonic:
        """Convert a list of harmonic records into a Harmonic object."""
        harmonic_block = [
            [harmonic.period, harmonic.amplitude, harmonic.phase]
            for harmonic in harmonics
        ]
        harmonic_model = Harmonic(
            name=user_defined_name,
            function="harmonic",
            quantityunitpair=[
                QuantityUnitPair(quantity="harmonic component", unit="minutes"),
                QuantityUnitPair(quantity=f"{quantity_name} amplitude", unit=unit),
                QuantityUnitPair(quantity=f"{quantity_name} phase", unit="deg"),
            ],
            datablock=harmonic_block,
        )
        return harmonic_model

    @staticmethod
    def convert_astronomic(
        user_defined_name,
        astronomics: List[AstronomicRecord],
        quantity_name: str,
        unit: str,
    ) -> Astronomic:
        """Convert a list of astronomic records into an Astronomic object."""
        astronomic_block = [
            [astronomic.name, astronomic.amplitude, astronomic.phase]
            for astronomic in astronomics
        ]
        astronomic_model = Astronomic(
            name=user_defined_name,
            function="astronomic",
            quantityunitpair=[
                QuantityUnitPair(quantity="astronomic component", unit="-"),
                QuantityUnitPair(quantity=f"{quantity_name} amplitude", unit=unit),
                QuantityUnitPair(quantity=f"{quantity_name} phase", unit="deg"),
            ],
            datablock=astronomic_block,
        )
        return astronomic_model


class TimToForcingConverter:
    """
    A class to convert TimModel data into ForcingModel data for boundary condition definitions.

    The class provides a static method `convert` to convert a TimModel object into a ForcingModel object.

    The method requires the following arguments:
    - `tim_model`: A TimModel object containing the time series data.
    - `start_time`: The reference time for the forcing data.
    - `time_interpolation`: The time interpolation method for the forcing data.
    - `units`: A list of units corresponding to the forcing quantities.
    - `user_defined_names`: A list of user-defined names for the forcing blocks.
    """

    @staticmethod
    def convert(
        tim_model: TimModel,
        time_unit: str,
        time_interpolation: str = "linear",
        units: List[str] = None,
        user_defined_names: List[str] = None,
    ) -> List[TimeSeries]:
        """
        Convert a TimModel into a ForcingModel.

        Args:
            tim_model (TimModel):
                The input TimModel to be converted.
            time_unit (str):
                Formatted string containing the units of time, including absolute datetime reference information
                (according to UDunits). For example, "minutes since 1992-10-8 15:15:42.5 -6:00".
            time_interpolation (str, optional):
                The time interpolation method for the forcing data. Defaults to "linear".
            units (List[str], optional):
                A list of units corresponding to the forcing quantities.
            user_defined_names (List[str], optional):
                A list of user-defined names for the forcing blocks.

        Returns:
            TimeSeries:
                The converted TimeSeries.

        Raises:
            ValueError: If `units` and `user_defined_names` are not provided.
            ValueError: If the lengths of `units`, `user_defined_names`, and the columns in the first row of the TimModel
                do not match.

        Examples:
            ```python
            >>> from hydrolib.core.dflowfm.tim.models import TimModel
            >>> from hydrolib.tools.extforce_convert.converters import TimToForcingConverter
            >>> file_path = "tests/data/input/tim/single_data_for_timeseries.tim"
            >>> user_defined_names = ["discharge"]
            >>> tim_model = TimModel(file_path, user_defined_names)
            >>> print(tim_model.as_dict())
            {'discharge': [0.0, 0.01, 0.0, -0.01, 0.0, 0.01, 0.0, -0.01, 0.0, 0.01, 0.0, -0.01, 0.0]}
            >>> converter = TimToForcingConverter()
            >>> time_series = converter.convert(
            ...     tim_model, "minutes since 2015-01-01 00:00:00", "linear", ["m3/s"], ["discharge"]
            ... )
            >>> print(time_series[0].name)
            discharge
            >>> print(time_series[0].datablock)
            [[0.0, 0.0], [10.0, 0.01], [20.0, 0.0], [30.0, -0.01], [40.0, 0.0], [50.0, 0.01], [60.0, 0.0], [70.0, -0.01], [80.0, 0.0], [90.0, 0.01], [100.0, 0.0], [110.0, -0.01], [120.0, 0.0]]

            ```
        """
        if units is None or user_defined_names is None:
            raise ValueError("Both 'units' and 'user_defined_names' must be provided.")

        if time_unit is None:
            raise ValueError("The 'start_time' must be provided.")

        first_record = tim_model.timeseries[0].data
        if len(units) != len(user_defined_names) != len(first_record):
            raise ValueError(
                "The lengths of 'units', 'user_defined_names' and length of the columns in the first row must match."
            )

        df = tim_model.as_dataframe()
        time_data = df.index.tolist()
        time_series_list = []
        for i, (column, vals) in enumerate(df.items()):
            unit = units[i]
            forcing = TimeSeries(
                name=user_defined_names[i],
                function="timeseries",
                timeinterpolation=time_interpolation,
                quantityunitpair=[
                    QuantityUnitPair(quantity="time", unit=time_unit),
                    QuantityUnitPair(quantity=column, unit=unit),
                ],
                datablock=[[i, j] for i, j in zip(time_data, vals.values.tolist())],
            )

            time_series_list.append(forcing)

        return time_series_list


class T3DToForcingConverter:
    """T3D to Forcing Converter."""

    @staticmethod
    def convert(
        t3d_models: List[T3DModel],
        quantities_names: List[str],
        user_defined_names: List[str] = None,
    ) -> List[T3D]:
        """Convert a list of T3DModel into a list of T3D Forcing to be saved into the .bc file."""
        t3d_forcings = []
        for label, model in zip(user_defined_names, t3d_models):
            model.quantities_names = quantities_names
            t3d = T3DToForcingConverter.convert_t3d_model(model, label)
            t3d_forcings.append(t3d)

        return t3d_forcings

    @staticmethod
    def convert_t3d_model(
        t3d_model: T3DModel,
        user_defined_name: List[str] = None,
    ) -> T3D:
        """Convert a T3DModel into a T3D Forcing to be saved into the .bc file.

        Args:
            t3d_model(T3DModel):
                T3DModel representing the .t3d file model.
            user_defined_name (List[str], optional):
                user-defined name for the forcing block.

        Returns:
            T3D: The converted T3D object.

        Examples:
            ```python
            >>> from hydrolib.core.dflowfm.t3d.models import T3DModel, T3DTimeRecord
            >>> from hydrolib.tools.extforce_convert.converters import T3DToForcingConverter
            >>> from hydrolib.core.dflowfm.bc.models import QuantityUnitPair, T3D
            >>> t3d_model = T3DModel(
            ...     layer_type="SIGMA",
            ...     layers=[0.0, 0.1, 0.2, 0.3, 0.4],
            ...     records = [
            ...         T3DTimeRecord(time="0 seconds since 2006-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10.0, 10.0]),
            ...         T3DTimeRecord(time="1e9 seconds since 2001-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10.0, 10.0])
            ...     ],
            ...     quantities_names=["temperature", "salinity", "discharge", "any quantity"]
            ... )
            >>> converter = T3DToForcingConverter()
            >>> t3d_forcing = converter.convert_t3d_model(t3d_model,"sigma-5-layers-time-steps")
            >>> print(t3d_forcing.name)
            sigma-5-layers-time-steps
            >>> print(t3d_forcing.function)
            t3d
            >>> print(t3d_forcing.datablock)
            [[0.0, 5.0, 5.0, 10.0, 10.0], [1000000000.0, 5.0, 5.0, 10.0, 10.0]]
            >>> print(t3d_forcing.quantityunitpair) # doctest: +SKIP
            [
                QuantityUnitPair(quantity='time', unit='seconds since 2006-01-01 00:00:00 +00:00', vertpositionindex=None),
                QuantityUnitPair(quantity='temperature', unit='degC', vertpositionindex=1),
                QuantityUnitPair(quantity='salinity', unit='ppt', vertpositionindex=2),
                QuantityUnitPair(quantity='discharge', unit='m3/s', vertpositionindex=3)
            ]
            >>> print(t3d_forcing.vertpositions)
            [0.0, 0.1, 0.2, 0.3, 0.4]
            >>> print(t3d_forcing.vertpositiontype)
            percBed

            ```
        """
        data = {
            "name": user_defined_name,
            "function": "t3d",
            "vertpositions": t3d_model.layers,
            "vertpositiontype": (
                "percBed"
                if not hasattr(t3d_model, "vertpositiontype")
                else t3d_model.vertpositiontype
            ),
        }

        if hasattr(t3d_model, "vertinterpolation"):
            data["vertinterpolation"] = t3d_model.vertinterpolation

        if hasattr(t3d_model, "timeinterpolation"):
            data["timeinterpolation"] = t3d_model.timeinterpolation

        data_dict = t3d_model.as_dict()
        updated = [[k] + v for k, v in data_dict.items()]
        data["datablock"] = updated

        time_unit = t3d_model.records[0].time_unit
        ref_date = t3d_model.records[0].reference_date
        quantities_list = [
            QuantityUnitPair(quantity="time", unit=f"{time_unit} since {ref_date}")
        ]

        units = t3d_model.get_units()
        quantities_names = t3d_model.quantities_names
        for i, (quantity, unit) in enumerate(zip(quantities_names, units)):
            quantities_list.append(
                QuantityUnitPair(quantity=quantity, unit=unit, vertpositionindex=i + 1)
            )

        data["quantityunitpair"] = quantities_list

        t3d = T3D(**data)
        return t3d



"""models.py defines all classes and functions related to representing pol/pli(z) files.
"""

from typing import Callable, List, Optional, Sequence, Tuple

from pydantic import Field

from hydrolib.core.base.models import BaseModel, ModelSaveSettings, ParsableFileModel


class Description(BaseModel):
    """Description of a single PolyObject.

    The Description will be prepended to a block. Each line will
    start with a '*'.

    Attributes:
        content (str): The content of this Description.
    """

    content: str


class Metadata(BaseModel):
    """Metadata of a single PolyObject.

    Attributes:
        name (str): The name of the PolyObject
        n_rows (int): The number of rows (i.e. Point instances) of the PolyObject
        n_columns (int): The total number of values in a Point, including x, y, and z.
    """

    name: str
    n_rows: int
    n_columns: int


class Point(BaseModel):
    """Point consisting of a x and y coordinate, an optional z coordinate and data.

    Attributes:
        x (float): The x-coordinate of this Point
        y (float): The y-coordinate of this Point
        z (Optional[float]): An optional z-coordinate of this Point.
        data (Sequence[float]): The additional data variables of this Point.
    """

    x: float
    y: float
    z: Optional[float] = None
    data: Sequence[float]

    def _get_identifier(self, data: dict) -> Optional[str]:
        x = data.get("x")
        y = data.get("y")
        z = data.get("z")
        return f"x:{x} y:{y} z:{z}"


class PolyObject(BaseModel):
    """PolyObject describing a single block in a poly file.

    The metadata should be consistent with the points:
    - The number of points should be equal to number of rows defined in the metadata
    - The data of each point should be equal to the number of columns defined in the
      metadata.

    Attributes:
        description (Optional[Description]):
            An optional description of this PolyObject
        metadata (Metadata):
            The Metadata of this PolObject, describing the structure
        points (List[Point]):
            The points describing this PolyObject, structured according to the Metadata
    """

    description: Optional[Description] = None
    metadata: Metadata
    points: List[Point]


class PolyFile(ParsableFileModel):
    """
    Poly-file (.pol/.pli/.pliz) representation.

    Notes:
        - The `has_z_values` attribute is used to determine if the PolyFile contains z-values.
        - The `has_z_values` is false by default and should be set to true if the PolyFile path ends with `.pliz`.
        - The `***.pliz` file should have a 2*3 structure, where the third column contains the z-values, otherwise
        (the parser will give an error).
        - If there is a label in the file, the parser will ignore the label and read the file as a normal polyline file.
        ```
        tfl_01
            2 2
            0.00 1.00 #zee
            0.00 2.00 #zee
        ```
        - if the file is .pliz, and the dimensions are 2*5 the first three columns will be considered as x, y, z values
        and the last two columns will be considered as data values.
        ```
        L1
            2 5
            63.35 12.95 -4.20 -5.35 0
            45.20 6.35 -3.00 -2.90 0
        ```
    """

    has_z_values: bool = False
    objects: Sequence[PolyObject] = Field(default_factory=list)

    def _serialize(self, _: dict, save_settings: ModelSaveSettings) -> None:
        from .serializer import write_polyfile

        # We skip the passed dict for a better one.
        write_polyfile(self._resolved_filepath, self.objects, self.serializer_config)

    @classmethod
    def _ext(cls) -> str:
        return ".pli"

    @classmethod
    def _filename(cls) -> str:
        return "objects"

    @classmethod
    def _get_serializer(cls) -> Callable:
        # Unused, but requires abstract implementation
        pass

    @classmethod
    def _get_parser(cls) -> Callable:
        # Prevent circular dependency in Parser
        from hydrolib.core.dflowfm.polyfile.parser import read_polyfile

        return read_polyfile

    @property
    def x(self) -> List[float]:
        """X-coordinates of all points in the PolyFile."""
        return [point.x for obj in self.objects for point in obj.points]

    @property
    def y(self) -> List[float]:
        """Y-coordinates of all points in the PolyFile."""
        return [point.y for obj in self.objects for point in obj.points]

    def get_z_sources_sinks(self) -> Tuple[List[float], List[float]]:
        """
        Get the z values of the source and sink points from the polyline file.

        Returns:
            z_source, z_sinkA: Tuple[List[float]]:
            If the polyline has data (more than 3 columns), then both the z_source and z_sink will be a list of two values.
            Otherwise, the z_source and the z_sink will be a single value each.

        Note:
             - calling this method on a polyline file that does not have z-values will return a list of None.

        Examples:
            - in case the polyline has 3 columns:
                ```python
                >>> polyline = PolyFile("tests/data/input/source-sink/leftsor.pliz")
                >>> z_source, z_sink = polyline.get_z_sources_sinks()
                >>> print(z_source, z_sink)
                [-3.0] [-4.2]

                ```
            - in case the polyline has more than 3 columns:
                ```python
                >>> polyline = PolyFile("tests/data/input/source-sink/leftsor-5-columns.pliz") #Doctest: +SKIP
                >>> z_source, z_sink = polyline.get_z_sources_sinks()
                >>> print(z_source, z_sink)
                [-3.0, -2.9] [-4.2, -5.35]

                ```
            - in case the polyline does not have z-values:
                ```python
                >>> root_dir = "tests/data/input/dflowfm_individual_files/polylines"
                >>> polyline = PolyFile(f"{root_dir}/boundary-polyline-no-z-no-label.pli")
                >>> z_source, z_sink = polyline.get_z_sources_sinks()
                >>> print(z_source, z_sink)
                [None] [None]

                ```
        """
        has_data = True if self.objects[0].points[0].data else False

        z_source_sink = []
        for elem in [0, -1]:
            point = self.objects[0].points[elem]
            if has_data:
                z_source_sink.append([point.z, point.data[0]])
            else:
                z_source_sink.append([point.z])

        z_sink: list[float | None] = z_source_sink[0]
        z_source: list[float | None] = z_source_sink[1]
        return z_source, z_sink

    @property
    def number_of_points(self) -> int:
        """Total number of points in the PolyFile."""
        return len(self.x)



"""Converter for old external forcing files to the new format."""

import os
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

from tqdm import tqdm

from hydrolib.core.base.file_manager import PathOrStr
from hydrolib.core.base.utils import PathStyle
from hydrolib.core.dflowfm.ext.models import (
    Boundary,
    ExtModel,
    Lateral,
    Meteo,
    SourceSink,
)
from hydrolib.core.dflowfm.extold.models import ExtOldModel
from hydrolib.core.dflowfm.inifield.models import (
    IniFieldModel,
    InitialField,
    ParameterField,
)
from hydrolib.core.dflowfm.structure.models import Structure, StructureModel
from hydrolib.tools.extforce_convert.converters import (
    BoundaryConditionConverter,
    ConverterFactory,
    InitialConditionConverter,
    ParametersConverter,
    SourceSinkConverter,
)
from hydrolib.tools.extforce_convert.mdu_parser import MDUParser
from hydrolib.tools.extforce_convert.utils import (
    CONVERTER_DATA,
    backup_file,
    construct_filemodel_new_or_existing,
    path_relative_to_parent,
)


class ExternalForcingConverter:
    """Converter for old external forcing files to the new format."""

    def __init__(
        self,
        extold_model: Union[PathOrStr, ExtOldModel],
        ext_file: Optional[PathOrStr] = None,
        inifield_file: Optional[PathOrStr] = None,
        structure_file: Optional[PathOrStr] = None,
        mdu_parser: MDUParser = None,
        verbose: bool = False,
        path_style: PathStyle = None,
        debug: Optional[bool] = False,
    ):
        r"""Initialize the converter.

        The converter will create new external forcing, initial field and structure files in the same directory as the
        old external forcing file, if no paths were given by the user for the new models.

        Args:
            extold_model (PathOrStr, ExtOldModel):
                ExtOldModel or path to the old external forcing file. if `ExtOldModel` without a filepath the
                `ExtOldModel` will be assumed to be in the, current working directory.
            ext_file (PathOrStr, optional):
                Path to the new external forcing file.
            inifield_file (PathOrStr, optional):
                Path to the initial field file.
            structure_file (PathOrStr, optional):
                Path to the structure file.
            mdu_parser (Optional[MDUParser], optional):
                a Parser for the MDU file.
            verbose (bool, optional, Defaults to False):
                Enable verbose output.
            path_style (Optional[PathStyle], optional):
                Specifies how absolute paths in model files are interpreted and converted (unix or windows style).
                Use this to correctly handle Unix paths on Windows or Windows paths on Unix.
                If None, the system's default style is used.
                - Example (Windows): `P:/path/to/file`
                - Example (Unix): `/p/path/to/file`
            debug (bool, Optional):
                Enable debug mode. In debug mode unsupported quantities will be skipped and not raise an error.
                Defaults to False.

        Raises:
            FileNotFoundError: If the old external forcing file does not exist.

        Examples:
            - Create a converter from old external forcing object:
            ```python
            >>> from hydrolib.core.dflowfm.extold.models import ExtOldModel
            >>> root_dir = Path("path/to/your/root/dir")
            >>> forcing_data = {
            ...     'QUANTITY': 'windspeedfactor',
            ...     'FILENAME': rf'{root_dir}/my-poly-file.pol',
            ...     'FILETYPE': '11',
            ...     'METHOD': '4',
            ...     'OPERAND': 'O',
            ... }
            >>> forcing_model_data = {
            ... 'comment': [' Example (old-style) external forcings file'],
            ... 'forcing': [forcing_data]
            ... }
            >>> old_model = ExtOldModel(**forcing_model_data) #doctest: +SKIP
            >>> converter = ExternalForcingConverter(extold_model=old_model) #doctest: +SKIP

            ```
            - Create a converter from an old external forcing file:
            ```python
            >>> converter = ExternalForcingConverter("old-external-forcing.ext") #doctest: +SKIP
            >>> converter.update() #doctest: +SKIP
            ```
        """
        if isinstance(extold_model, Path) or isinstance(extold_model, str):
            # if the extold model is given as path/str then read the file
            extold_model = self._read_old_file(extold_model, path_style=path_style)
        else:
            if not isinstance(extold_model, ExtOldModel):
                raise TypeError(
                    "extold_model must be a PathOrStr or ExtOldModel instance."
                )

        # if the mdu is not given then the root dir is the same as the extold model
        if mdu_parser is not None:
            rdir = mdu_parser.mdu_path.parent
        else:
            rdir = extold_model.filepath.parent

        self._extold_model = extold_model
        self._verbose = verbose
        self._root_dir = rdir
        self._path_style = path_style

        # create the new models if not provided by the user in the same directory as the old external file
        path = rdir / "new-external-forcing.ext" if ext_file is None else ext_file
        self._ext_model = construct_filemodel_new_or_existing(
            ExtModel, path, recurse=False
        )

        path = (
            rdir / "new-initial-conditions.ini"
            if inifield_file is None
            else inifield_file
        )
        self._inifield_model = construct_filemodel_new_or_existing(
            IniFieldModel, path, recurse=False
        )

        path = rdir / "new-structure.ini" if structure_file is None else structure_file
        self._structure_model = construct_filemodel_new_or_existing(
            StructureModel, path, recurse=False
        )

        if mdu_parser is not None:
            self.temperature_salinity_data: Dict[str, int] = (
                mdu_parser.temperature_salinity_data
            )
            self._mdu_parser = mdu_parser

        self._legacy_files = []
        self.debug = debug
        self.un_supported_quantities = self.check_unsupported_quantities()

    def check_unsupported_quantities(self):
        """Check for unsupported quantities in the old external forcing model.

        Returns:
            list[str]: List of unsupported quantities found in the old external forcing model.
        """
        # check if the file contains unsupported quantities
        return CONVERTER_DATA.check_unsupported_quantities(
            self.extold_model, raise_error=not self.debug
        )

    @property
    def legacy_files(self) -> list[Path]:
        """List of legacy files that were created during the conversion."""
        return self._legacy_files

    @legacy_files.setter
    def legacy_files(self, value: list[Path]):
        """Set the legacy files."""
        if not isinstance(value, list):
            raise TypeError("legacy_files must be a list of Path objects.")
        self._legacy_files += value

    @property
    def verbose(self) -> bool:
        """bool: Enable verbose output."""
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool):
        self._verbose = value

    @property
    def mdu_parser(self) -> MDUParser:
        if hasattr(self, "_mdu_parser"):
            val = self._mdu_parser
        else:
            val = None
        return val

    @property
    def root_dir(self) -> Path:
        """Root directory of the external forcing file."""
        return self._root_dir

    @property
    def path_style(self) -> PathStyle:
        """Path style for the absolute paths in the models."""
        return self._path_style

    @property
    def extold_model(self) -> ExtOldModel:
        """Old external forcing model."""
        return self._extold_model

    @property
    def ext_model(self) -> ExtModel:
        """New External forcing Model."""
        if not hasattr(self, "_ext_model"):
            raise ValueError(
                "ext_model not set, please use the `ext_model` setter. to set it."
            )
        return self._ext_model

    @ext_model.setter
    def ext_model(self, path: PathOrStr):
        """New external forcing model.

        Args:
            path (PathOrStr, optional): Path to the new external forcing file.
        """
        self._ext_model = construct_filemodel_new_or_existing(ExtModel, path)

    @property
    def inifield_model(self) -> IniFieldModel:
        """IniFieldModel: object with all initial fields blocks."""
        if not hasattr(self, "_inifield_model"):
            raise ValueError(
                "inifield_model not set, please use the `inifield_model` setter. to set it."
            )
        return self._inifield_model

    @inifield_model.setter
    def inifield_model(self, path: PathOrStr):
        self._inifield_model = construct_filemodel_new_or_existing(IniFieldModel, path)

    @property
    def structure_model(self) -> StructureModel:
        """StructureModel: object with all structure blocks."""
        if not hasattr(self, "_structure_model"):
            raise ValueError(
                "structure_model not set, please use the `structure_model` setter. to set it."
            )
        return self._structure_model

    @structure_model.setter
    def structure_model(self, path: PathOrStr):
        self._structure_model = construct_filemodel_new_or_existing(
            StructureModel, path
        )

    @staticmethod
    def _read_old_file(
        ext_old_file: PathOrStr, path_style: Optional[PathStyle] = None
    ) -> ExtOldModel:
        """Read a legacy D-Flow FM external forcings file (.ext) into an ExtOldModel object.

        Args:
            ext_old_file (PathOrStr):
                path to the external forcings file (.ext)
            path_style (Optional[PathStyle]):
                Path style for the file paths in the models. If None, the system style is used.
                Converts absolute paths based on the provided style to the system style.

            Returns:
                ExtOldModel: object with all forcing blocks.

            Raises:
                FileNotFoundError: If the old external forcing file does not exist.
        """
        if not isinstance(ext_old_file, Path):
            ext_old_file = Path(ext_old_file)

        if not ext_old_file.exists():
            raise FileNotFoundError(f"File not found: {ext_old_file}")

        ext_old_model = ExtOldModel(ext_old_file, path_style=path_style)

        return ext_old_model

    def _type_field_map(self) -> dict[type, tuple[Any, str]]:
        return {
            Boundary: (self.ext_model, "boundary"),
            Lateral: (self.ext_model, "lateral"),
            SourceSink: (self.ext_model, "sourcesink"),
            Meteo: (self.ext_model, "meteo"),
            InitialField: (self.inifield_model, "initial"),
            ParameterField: (self.inifield_model, "parameter"),
            Structure: (self.structure_model, "structure"),
        }

    def update(
        self,
    ) -> Union[Tuple[ExtModel, IniFieldModel, StructureModel], None]:
        """Convert the old external forcing file to a new format files.

        Notes:
            - When the output files exist, output will be appended to them.
            - If there is a new external forcing file, the converted quantities will be appended to it, otherwise a new
            external forcing file will be created.
            - If there is an initial field file, the converted quantities will be appended to it, otherwise a new initial
            field file will be created in the same directory as the mdu file, and the `IniFieldFile` field will be
            added/updated in the geometry section in the mdu file.

        Returns:
            Tuple[ExtOldModel, ExtModel, IniFieldModel, StructureModel]:
                The updated models (already written to disk). Maybe used
                at call site to inspect the updated models.
        """
        self._log_conversion_details()
        num_quantities = len(self.extold_model.forcing)

        type_field_map = self._type_field_map()

        with tqdm(
            total=num_quantities, desc="Converting forcings", unit="forcing"
        ) as progress_bar:
            for forcing in self.extold_model.forcing:
                if forcing.quantity.lower() in self.un_supported_quantities:
                    print(
                        f"The quantity {forcing.quantity} is not supported, and the debug mode is {self.debug}. "
                        "So the forcing will not be converted (stay in the old external forcing file)."
                    )
                    progress_bar.update(1)
                    continue

                # Update the description of tqdm to include the current forcing's filepath
                progress_bar.set_description(
                    f"Processing: {forcing.quantity} - {forcing.filename.filepath}"
                )

                new_quantity_block = self._convert_forcing(forcing)
                model_field = type_field_map.get(type(new_quantity_block))

                if model_field is None:
                    raise NotImplementedError(
                        f"Unsupported model class {type(new_quantity_block)} for {forcing.quantity} in "
                        f"{self.extold_model.filepath}."
                    )

                model, attr = model_field
                setattr(model, attr, getattr(model, attr) + [new_quantity_block])

                progress_bar.update(1)

        if self.mdu_parser is not None:
            self._update_mdu_file()

        if self.un_supported_quantities:
            # replace the forcings by the forcings that were not converted (the unsupported quantities)
            forcing = [
                forcing
                for forcing in self.extold_model.forcing
                if forcing.quantity.lower() in self.un_supported_quantities
            ]
            self.extold_model.forcing = forcing

        return self.ext_model, self.inifield_model, self.structure_model

    def _convert_forcing(self, forcing) -> Union[Boundary, Lateral, Meteo, SourceSink]:
        """Convert a single forcing block to the appropriate new format.

        Notes:
            - The SourceSink converter needs the salinity and temperature from the FM model.
            - The BoundaryCondition converter needs the start time from the FM model.
        """
        converter_class = ConverterFactory.create_converter(forcing.quantity)
        converter_class.root_dir = self.root_dir

        # only the SourceSink converter needs the quantities' list
        if isinstance(converter_class, SourceSinkConverter):

            if self.temperature_salinity_data is None:
                raise ValueError(
                    "FM model is required to convert SourcesSink quantities."
                )
            else:
                temp_salinity_mdu = self.temperature_salinity_data
                start_time = self.temperature_salinity_data.get("refdate")

            quantities = self.extold_model.quantities
            new_quantity_block = converter_class.convert(
                forcing, quantities, start_time=start_time, **temp_salinity_mdu
            )
        elif isinstance(converter_class, BoundaryConditionConverter):
            if self.temperature_salinity_data is None:
                raise ValueError("FM model is required to convert Boundary conditions.")
            else:
                start_time = self.temperature_salinity_data.get("refdate")
                new_quantity_block = converter_class.convert(forcing, start_time)
        elif isinstance(
            converter_class, (InitialConditionConverter, ParametersConverter)
        ):
            forcing_path = path_relative_to_parent(
                forcing,
                self.inifield_model.filepath,
                self.extold_model.filepath,
                self.mdu_parser,
            )
            new_quantity_block = converter_class.convert(forcing, forcing_path)
        else:
            new_quantity_block = converter_class.convert(forcing)

        if hasattr(converter_class, "legacy_files"):
            self.legacy_files = converter_class.legacy_files

        return new_quantity_block

    def save(self, backup: bool = True, recursive: bool = True):
        """Save the updated models to disk.

        Args:
            backup (bool, optional):
                Create a backup of each file that will be overwritten.
            recursive (bool, optional): Defaults to True.
                Save the models recursively.
        """
        num_quantities_inifield = len(self.inifield_model.parameter) + len(
            self.inifield_model.initial
        )
        if num_quantities_inifield > 0:
            self._save_inifield_model(backup, recursive)

        if len(self.structure_model.structure) > 0:
            self._save_structure_model(backup, recursive)

        if self.un_supported_quantities:
            backup_file(self.extold_model.filepath)
            self.extold_model.save(recurse=recursive, exclude_unset=True)

        num_quantities_ext = (
            len(self.ext_model.meteo)
            + len(self.ext_model.sourcesink)
            + len(self.ext_model.boundary)
            + len(self.ext_model.lateral)
        )
        if num_quantities_ext:
            if backup and self.ext_model.filepath.exists():
                backup_file(self.ext_model.filepath)
            self.ext_model.save(
                recurse=recursive, exclude_unset=True, path_style=self.path_style
            )

        if self.mdu_parser is not None:
            self.mdu_parser.clean()
            self.mdu_parser.save(backup=backup)

    def _save_inifield_model(self, backup: bool, recursive: bool):
        """
        The save method for the IniFieldModel.
        the exclude_unset parameter is set to False, to save the general section with the fileversion section as
        required by the Fortran kernel.
        """
        if backup and self.inifield_model.filepath.exists():
            backup_file(self.inifield_model.filepath)
        self.inifield_model.save(
            recurse=recursive, exclude_unset=False, path_style=self.path_style
        )

    def _save_structure_model(self, backup: bool, recursive: bool):
        if backup and self.structure_model.filepath.exists():
            backup_file(self.structure_model.filepath)
        self.structure_model.save(
            recurse=recursive, exclude_unset=True, path_style=self.path_style
        )

    def clean(self):
        """Clean the directory from the old external forcing file and the time file."""
        if len(self.legacy_files) > 0:
            for file in self.legacy_files:
                print(f"Removing legacy file:{file}")
                file.unlink()

        if not self.un_supported_quantities:
            self.extold_model.filepath.unlink()

    @classmethod
    def from_mdu(
        cls,
        mdu_file: PathOrStr,
        ext_file_user: Optional[PathOrStr] = None,
        inifield_file_user: Optional[PathOrStr] = None,
        structure_file_user: Optional[PathOrStr] = None,
        path_style: Optional[PathStyle] = None,
        debug: bool = False,
    ) -> "ExternalForcingConverter":
        """Create the converter from the MDU file.

        Args:
            mdu_file (PathOrStr): Path to the D-Flow FM main input file (.mdu).
                Must be parsable into a standard FMModel.
                When this contains a valid filename for ExtFile, conversion
                will be performed.
            ext_file_user (PathOrStr, optional): Path to the output external forcings
                file. Defaults to the given ExtForceFileNew in the MDU file, if
                present, or forcings.ext otherwise.
            inifield_file_user (PathOrStr, optional): Path to the output initial field
                file. Defaults to the given IniFieldFile in the MDU file, if
                present, or inifields.ini otherwise.
            structure_file_user (PathOrStr, optional): Path to the output structures.ini
                file. Defaults to the given StructureFile in the MDU file, if
                present, or structures.ini otherwise.
            path_style (Optional[PathStyle], optional):
                Path style for the file paths in the models. If None, the system style is used.
                Converts absolute paths based on the provided style to the system style.
            debug (bool, Optional):
                Enable debug mode. In debug mode unsupported quantities will be skipped and not raise an error.
                Defaults to False.

        Returns:
            ExternalForcingConverter: The converter object.

        Raises:
            FileNotFoundError: If the MDU file does not exist.
            ValueError: If the old external forcing file is not found in the MDU file.
            DeprecationWarning: If the MDU file contains unknown keywords.
        """
        mdu_parser = MDUParser(mdu_file)

        extoldfile = (
            mdu_parser.mdu_path.parent / mdu_parser.extforce_block.extforce_file
        )

        ext_file_user = mdu_parser.extforce_block.get_new_extforce_file(ext_file_user)
        inifield_file_user = mdu_parser.get_inifield_file(inifield_file_user)
        structure_file_user = mdu_parser.get_structure_file(structure_file_user)

        return cls(
            extoldfile,
            ext_file_user,
            inifield_file_user,
            structure_file_user,
            mdu_parser,
            path_style=path_style,
            debug=debug,
        )

    def _update_mdu_file(self):
        """Update the MDU file with the new external forcings files, initial fields and structures files.

        Notes:
            - The MDU file is updated in place.
            - The InifieldFile and StructureFile blocks are updated only if they were not present in the file,
            if not, they will be created in the same directory as the mdu file.
        """
        num_ext_model_quantities = (
            len(self.ext_model.boundary)
            + len(self.ext_model.lateral)
            + len(self.ext_model.meteo)
            + len(self.ext_model.sourcesink)
        )

        remove_old_ext_file = False if self.un_supported_quantities else True

        self.mdu_parser.update_extforce_file_new(
            self.ext_model.filepath.name,
            num_quantities=num_ext_model_quantities,
            remove_old_ext_file=remove_old_ext_file,
        )

        num_quantities_inifield = len(self.inifield_model.parameter) + len(
            self.inifield_model.initial
        )

        if num_quantities_inifield > 0:
            self.mdu_parser.update_inifield_file(self.inifield_model.filepath.name)

        if len(self.structure_model.structure) > 0:
            self.mdu_parser.update_structure_file(self.structure_model.filepath.name)

    def _log_conversion_details(self):
        """Log details about the conversion process if verbosity is enabled."""
        if self.verbose:
            workdir = os.getcwd() + "\\"
            print(f"Work dir: {workdir}")
            print("Using attribute files:")
            print("Input:")
            print(f"* {self.extold_model.filepath}")
            print("Output:")
            print(f"* {self.ext_model.filepath}")
            print(f"* {self.inifield_model.filepath}")
            print(f"* {self.structure_model.filepath}")


def recursive_converter(
    root_dir: PathOrStr,
    backup: bool = True,
    suppress_errors: bool = False,
    remove_legacy: bool = False,
    debug: bool = False,
):
    """Migrate all external forcings files in a directory tree to the new format.

    Args:
        root_dir: Directory to recursively find and convert .mdu files in.
        backup (bool, optional): Create a backup of each file that will be overwritten.
        suppress_errors (bool, optional): Suppress errors during conversion.
        remove_legacy (bool, optional): Remove legacy/old files (e.g. .tim) after conversion. Defaults to False.
        debug (bool, Optional):
                Enable debug mode. In debug mode unsupported quantities will be skipped and not raise an error.
                Defaults to False.
    """
    mdu_files = [
        path for path in Path(root_dir).rglob("*.mdu") if "_ext" not in path.name
    ]

    if not mdu_files:
        print("No .mdu files found in the specified directory.")
    else:
        print(f"Found {len(mdu_files)} .mdu files. Starting conversion...")

        for path in tqdm(mdu_files, desc="Converting files"):
            try:
                converter = ExternalForcingConverter.from_mdu(path, debug=debug)
                _, _, _ = converter.update()
                converter.save(backup=backup)
                if remove_legacy:
                    converter.clean()
            except Exception as e:
                if not suppress_errors:
                    print(f"Error processing {path}: {e}")



"""T3D File Parser."""

from pathlib import Path
from typing import Any, Dict, List

from hydrolib.core.base.parser import BaseParser


class T3DParser(BaseParser):
    """
    A parser for .t3d files.
    Full line comments at the start of the file are supported. Comment lines start with either a `*` or a `#`.
    No other comments are supported.
    """

    @staticmethod
    def parse(filepath: Path) -> Dict[str, List[Any]]:
        """Parse a .t3d file into a dictionary with comments and parsed data from the t3d file.

        Args:
            filepath (Path): Path to the .t3d file to be parsed.

        Returns:
            Dict[str, List[Any]]: A dictionary with keys "records", "layers", and "layer_type".
            - "comments" represents comments found at the start of the file.
            - "layer_type": The type of the layer ["SIGMA", "Z"].
            - "layers": A list of floats representing the layers.
            - "records" is a list of dictionaries with the key as "time" and values as "data".
                - "time" is a time as a string.
                - "data" is data as a list of floats.

        Raises:
            ValueError: If the file contains a comment that is not at the start of the file.
            ValueError: If the data of the timeseries is empty.

        Examples:
            >>> from pathlib import Path
            >>> from hydrolib.core.dflowfm.t3d.parser import T3DParser
            >>> data = T3DParser.parse(Path("tests/data/input/dflowfm_individual_files/t3d/sigma-5-layers-3-times.t3d"))
            >>> print(data) # doctest: +SKIP
            {
                "comments": [],
                "layer_type": "SIGMA",
                "layers": [0.0, 0.2, 0.6, 0.8, 1.0],
                "records": [
                    {
                        "time": "0 seconds since 2006-01-01 00:00:00 +00:00",
                        "data": [1.0, 1.0, 1.0, 1.0, 1.0]
                    },
                    {
                        "time": "180 seconds since 2001-01-01 00:00:00 +00:00",
                        "data": [2.0, 2.0, 2.0, 2.0, 2.0]
                    },
                    {
                        "time": "9999999 seconds since 2001-01-01 00:00:00 +00:00",
                        "data": [3.0, 3.0, 3.0, 3.0, 3.0]
                    }
                ]
            }
        """
        with filepath.open(encoding="utf8") as file:
            lines = file.readlines()
            comments, start_timeseries_index = T3DParser._read_header_comments(lines)
            data = T3DParser._read_data(lines, start_timeseries_index)

        return {"comments": comments} | data

    @staticmethod
    def _read_data(
        lines: List[str], start_timeseries_index: int
    ) -> Dict[str, List[Any]]:
        parsed_data = {"records": []}
        i = start_timeseries_index

        while i < len(lines):
            line = lines[i].strip()

            if len(line) == 0:
                i += 1
                continue

            T3DParser._raise_error_if_contains_comment(line, i + 1)

            # 1. Detect LAYER_TYPE
            if line.startswith("LAYER_TYPE="):
                # everything after the '=' is the type (e.g., "SIGMA")
                parsed_data["layer_type"] = line.split("=", 1)[1].strip()

            # 2. Detect LAYERS
            elif line.startswith("LAYERS="):
                # everything after the '=' is a space-separated list of floats
                layer_str = line.split("=", 1)[1].strip()
                parsed_data["layers"] = [float(x) for x in layer_str.split()]

            # 3. Detect TIME + next line of data
            elif line.startswith("TIME"):
                # Grab everything after the '='
                time_value = line.split("=", 1)[1].strip()
                i += 1
                # The next line should contain the data
                data_line = lines[i]
                data_values = [float(x) for x in data_line.split()]
                # Save into the dictionary
                parsed_data["records"].append({"time": time_value, "data": data_values})

            i += 1

        return parsed_data



"""Cross section models for D-Flow FM."""

import logging
from typing import Annotated, List, Literal, Optional, Union

from pydantic import (
    BeforeValidator,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)

from hydrolib.core.dflowfm.friction.models import FrictionType
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    LocationValidationConfiguration,
    LocationValidationFieldNames,
    UnknownKeywordErrorManager,
    enum_value_parser,
    make_list,
    split_string_on_delimiter,
    validate_correct_length,
    validate_location_specification,
)

logger = logging.getLogger(__name__)

frictionid_description = 'Name of the roughness variable associated with \
    this cross section. Either this parameter or \
    frictionType should be specified. If neither \
    parameter is specified, the frictionId defaults \
    to "Main".'

frictiontype_description = "Roughness type associated with this cross section \
    Either this parameter or frictionId should be specified."

frictionvalue_description = "Roughness value; its meaning depends on the roughness type selected \
    (only used if frictionType specified)."


class CrossDefGeneral(INIGeneral):
    """The crosssection definition file's `[General]` section with file meta data."""

    fileversion: str = Field("3.00", alias="fileVersion")
    filetype: Literal["crossDef"] = Field("crossDef", alias="fileType")


class CrossLocGeneral(INIGeneral):
    """The crosssection location file's `[General]` section with file meta data."""

    fileversion: str = Field("3.00", alias="fileVersion")
    filetype: Literal["crossLoc"] = Field("crossLoc", alias="fileType")


class CrossSectionDefinition(INIBasedModel):
    """CrossSectionDefinition.

    A `[Definition]` block for use inside a crosssection definition file,
    i.e., a [CrossDefModel][hydrolib.core.dflowfm.crosssection.models.CrossDefModel].

    This class is intended as an abstract class: various subclasses should
    define they actual types of crosssection definitions.
    """

    class Comments(INIBasedModel.Comments):
        """Comments for the CrossSectionDefinition class."""

        id: Optional[str] = "Unique cross-section definition id."
        thalweg: Optional[str] = Field(
            "Transverse Y coordinate at which the cross section aligns with the branch (Keyword used by GUI only)."
        )

    comments: Comments = Comments()

    _header: Literal["Definition"] = "Definition"

    id: str = Field(alias="id")
    type: str = Field(alias="type")
    thalweg: Optional[float] = None

    @classmethod
    def _get_unknown_keyword_error_manager(cls) -> Optional[UnknownKeywordErrorManager]:
        """The CrossSectionDefinition does not currently support raising an error on unknown keywords."""
        return None

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id")

    @classmethod
    def _duplicate_keys_as_list(cls):
        return True

    @staticmethod
    def _check_friction_fields(frictionid, frictiontype, frictionvalue, label=""):
        """Check the friction specification.

        Make a model_validator that verifies whether the crosssection definition (subclass)
        has a valid friction specification.
        Supposed to be embedded in subclasses for their friction fields.

        Args:
            frictionid:
                name of the frictionid attribute in the subclass.
            frictiontype:
                name of the frictiontype attribute in the subclass.
            frictionvalue:
                name of the frictionvalue attribute in the subclass.

        Raises:
            ValueError:
                If both frictionid and frictiontype/frictionvalue are specified.
        """
        if frictionid and (frictiontype or frictionvalue):
            raise ValueError(
                f"{label} has duplicate friction specification: both frictionid and frictiontype/frictionvalue."
            )


class CircleCrsDef(CrossSectionDefinition):
    """CircleCrsDef.

    Crosssection definition with `type=circle`, to be included in a crossdef file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile.definition[..]`

    All lowercased attributes match with the circle input as described in
    [UM Sec.C.16.1.1](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.16.1.1).
    """

    class Comments(CrossSectionDefinition.Comments):
        """Comments for the CircleCrsDef class."""

        type: Optional[str] = Field("Cross section type; must read circle")

        diameter: Optional[str] = Field("Internal diameter of the circle [m].")
        frictionid: Optional[str] = Field(
            frictionid_description,
            alias="frictionId",
        )
        frictiontype: Optional[str] = Field(
            frictiontype_description,
            alias="frictionType",
        )
        frictionvalue: Optional[str] = Field(
            frictionvalue_description,
            alias="frictionValue",
        )

    comments: Comments = Comments()

    type: Literal["circle"] = Field("circle")
    diameter: float
    frictionid: Optional[str] = Field(None, alias="frictionId")
    frictiontype: Optional[FrictionType] = Field(None, alias="frictionType")
    frictionvalue: Optional[float] = Field(None, alias="frictionValue")

    @model_validator(mode="after")
    def check_friction(self):
        self._check_friction_fields(
            self.frictionid, self.frictiontype, self.frictionvalue, label=self.id
        )
        return self

    @field_validator("frictiontype", mode="before")
    @classmethod
    def validate_enum_frictiontype(cls, v):
        return enum_value_parser(v, FrictionType)


class RectangleCrsDef(CrossSectionDefinition):
    """RectangleCrsDef.

    Crosssection definition with `type=rectangle`, to be included in a crossdef file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile.definition[..]`

    All lowercased attributes match with the rectangle input as described in
    [UM Sec.C.16.1.2](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.16.1.2).
    """

    class Comments(CrossSectionDefinition.Comments):
        """Comments for the RectangleCrsDef class."""

        type: Optional[str] = Field("Cross section type; must read rectangle")
        width: Optional[str] = Field("Width of the rectangle [m].")
        height: Optional[str] = Field("Height of the rectangle [m].")
        closed: Optional[str] = Field("no: Open channel, yes: Closed channel.")
        frictionid: Optional[str] = Field(
            frictionid_description,
            alias="frictionId",
        )
        frictiontype: Optional[str] = Field(
            frictiontype_description,
            alias="frictionType",
        )
        frictionvalue: Optional[str] = Field(
            frictionvalue_description,
            alias="frictionValue",
        )

    comments: Comments = Comments()

    type: Literal["rectangle"] = Field("rectangle")
    width: float
    height: float
    closed: bool = Field(True)
    frictionid: Optional[str] = Field(None, alias="frictionId")
    frictiontype: Optional[FrictionType] = Field(None, alias="frictionType")
    frictionvalue: Optional[float] = Field(None, alias="frictionValue")

    @model_validator(mode="after")
    def check_friction(self):
        self._check_friction_fields(
            self.frictionid, self.frictiontype, self.frictionvalue, label=self.id
        )
        return self

    @field_validator("frictiontype", mode="before")
    @classmethod
    def validate_enum_frictiontype(cls, v):
        return enum_value_parser(v, FrictionType)


class ZWRiverCrsDef(CrossSectionDefinition):
    """ZWRiverCrsDef.

    Crosssection definition with `type=zwRiver`, to be included in a crossdef file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile.definition[..]`

    All lowercased attributes match with the zwRiver input as described in
    [UM Sec.C.16.1.3](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.16.1.3).
    """

    class Comments(CrossSectionDefinition.Comments):
        """Comments for the ZWRiverCrsDef class."""

        type: Optional[str] = Field(
            "Cross section type; must read zwRiver", alias="type"
        )
        numlevels: Optional[str] = Field(
            "Number of levels in the table.", alias="numLevels"
        )
        levels: Optional[str] = Field(
            "Space separated list of monotonic increasing heights/levels [m AD].",
            alias="levels",
        )
        flowwidths: Optional[str] = Field(
            "Space separated list of flow widths at the selected heights [m)].",
            alias="flowWidths",
        )
        totalwidths: Optional[str] = Field(
            "Space separated list of total widths at the selected heights [m]. "
            "Equal to flowWidths if not specified. If specified, the totalWidths"
            "should be larger than flowWidths.",
            alias="totalWidths",
        )
        leveecrestLevel: Optional[str] = Field(
            "Crest level of levee [m AD].", alias="leveeCrestlevel"
        )
        leveebaselevel: Optional[str] = Field(
            "Base level of levee [m AD].", alias="leveeBaseLevel"
        )
        leveeflowarea: Optional[str] = Field(
            "Flow area behind levee [m2].", alias="leveeFlowArea"
        )
        leveetotalarea: Optional[str] = Field(
            "Total area behind levee [m2].", alias="leveeTotalArea"
        )
        mainwidth: Optional[str] = Field(
            "Width of main section [m]. Default value: max(flowWidths).",
            alias="mainWidth",
        )
        fp1width: Optional[str] = Field(
            "Width of floodplain 1 section [m]. Default value: max(flowWidths)-mainWidth",
            alias="fp1Width",
        )
        fp2width: Optional[str] = Field(
            "Width of floodplain 2 section [m]. Default value: max(flowWidths)-mainWidth-fp1Width",
            alias="fp2Width",
        )
        frictionids: Optional[str] = Field(
            "Semicolon separated list of roughness variable names associated with the roughness "
            "sections. Either this parameter or frictionTypes should be specified. If neither "
            'parameter is specified, the frictionIds default to "Main", "FloodPlain1" '
            'and "FloodPlain2".',
            alias="frictionIds",
        )
        frictiontypes: Optional[str] = Field(
            "Semicolon separated list of roughness types associated with the roughness sections. "
            "Either this parameter or frictionIds should be specified. Can be specified as a "
            "single value if all roughness sections use the same type.",
            alias="frictionTypes",
        )
        frictionvalues: Optional[str] = Field(
            "Space separated list of roughness values; their meaning depends on the roughness "
            "types selected (only used if frictionTypes specified).",
            alias="frictionValues",
        )

    comments: Comments = Comments()

    type: Literal["zwRiver"] = Field("zwRiver")
    numlevels: int = Field(alias="numLevels")
    levels: List[float]
    flowwidths: List[float] = Field(alias="flowWidths")
    totalwidths: Optional[List[float]] = Field(None, alias="totalWidths")
    leveecrestLevel: Optional[float] = Field(None, alias="leveeCrestlevel")
    leveebaselevel: Optional[float] = Field(None, alias="leveeBaseLevel")
    leveeflowarea: Optional[float] = Field(None, alias="leveeFlowArea")
    leveetotalrea: Optional[float] = Field(None, alias="leveeTotalArea")
    mainwidth: Optional[float] = Field(None, alias="mainWidth")
    fp1width: Optional[float] = Field(None, alias="fp1Width")
    fp2width: Optional[float] = Field(None, alias="fp2Width")
    frictionids: Optional[List[str]] = Field(None, alias="frictionIds", delimiter=";")
    frictiontypes: Optional[List[FrictionType]] = Field(
        None, alias="frictionTypes", delimiter=";"
    )
    frictionvalues: Optional[List[float]] = Field(None, alias="frictionValues")

    @field_validator(
        "levels",
        "flowwidths",
        "totalwidths",
        "frictionvalues",
        "frictionids",
        "frictiontypes",
        mode="before",
    )
    @classmethod
    def split_string_list(cls, v, info: ValidationInfo):
        return split_string_on_delimiter(cls, v, info)

    @model_validator(mode="after")
    def check_friction(self):
        self._check_friction_fields(
            self.frictionids, self.frictiontypes, self.frictionvalues, label=self.id
        )
        return self

    @field_validator("frictiontypes", mode="before")
    @classmethod
    def validate_enum_frictiontype(cls, v):
        return enum_value_parser(v, FrictionType)

    @model_validator(mode="after")
    def check_list_lengths(self):
        """Validate that the length of the levels, flowwidths and totalwidths fields are as expected."""
        validate_correct_length(
            self.model_dump(),
            "levels",
            "flowwidths",
            "totalwidths",
            length_name="numlevels",
        )
        return self


class ZWCrsDef(CrossSectionDefinition):
    """ZWCrsDef.

    Crosssection definition with `type=zw`, to be included in a crossdef file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile.definition[..]`

    All lowercased attributes match with the zw input as described in
    [UM Sec.C.16.1.4](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.16.1.4).
    """

    class Comments(CrossSectionDefinition.Comments):
        """Comments for the ZWCrsDef class."""

        type: Optional[str] = Field("Cross section type; must read zw", alias="type")
        # NOTE: Field "template" deliberately ignored for now.
        numlevels: Optional[str] = Field(
            "Number of levels in the table.", alias="numLevels"
        )
        levels: Optional[str] = Field(
            "Space separated list of monotonic increasing heights/levels [m AD].",
            alias="levels",
        )
        flowwidths: Optional[str] = Field(
            "Space separated list of flow widths at the selected heights [m)].",
            alias="flowWidths",
        )
        totalwidths: Optional[str] = Field(
            "Space separated list of total widths at the selected heights [m]. "
            "Equal to flowWidths if not specified. If specified, the totalWidths"
            "should be larger than flowWidths.",
            alias="totalWidths",
        )
        frictionid: Optional[str] = Field(
            frictionid_description,
            alias="frictionId",
        )
        frictiontype: Optional[str] = Field(
            frictiontype_description,
            alias="frictionType",
        )
        frictionvalue: Optional[str] = Field(
            frictionvalue_description,
            alias="frictionValue",
        )

    comments: Comments = Comments()

    type: Literal["zw"] = Field("zw")
    numlevels: int = Field(alias="numLevels")
    levels: List[float]
    flowwidths: List[float] = Field(alias="flowWidths")
    totalwidths: Optional[List[float]] = Field(None, alias="totalWidths")
    frictionid: Optional[str] = Field(None, alias="frictionId")
    frictiontype: Optional[FrictionType] = Field(None, alias="frictionType")
    frictionvalue: Optional[float] = Field(None, alias="frictionValue")

    @field_validator(
        "levels",
        "flowwidths",
        "totalwidths",
        mode="before",
    )
    @classmethod
    def split_string_list(cls, v, info: ValidationInfo):
        return split_string_on_delimiter(cls, v, info)

    @model_validator(mode="after")
    def check_list_lengths(self):
        """Validate that the length of the levels, flowwidths and totalwidths fields are as expected."""
        validate_correct_length(
            self.model_dump(),
            "levels",
            "flowwidths",
            "totalwidths",
            length_name="numlevels",
        )
        return self

    @model_validator(mode="after")
    def check_friction(self):
        self._check_friction_fields(
            self.frictionid, self.frictiontype, self.frictionvalue, label=self.id
        )
        return self

    @field_validator("frictiontype", mode="before")
    @classmethod
    def validate_enum_frictiontype(cls, v):
        return enum_value_parser(v, FrictionType)


class YZCrsDef(CrossSectionDefinition):
    """YZCrsDef.

    Crosssection definition with `type=yz`, to be included in a crossdef file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile.definition[..]`

    All lowercased attributes match with the yz input as described in
    [UM Sec.C.16.1.6](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.16.1.6).
    """

    class Comments(CrossSectionDefinition.Comments):
        """Comments for the YZCrsDef class."""

        type: Optional[str] = Field("Cross section type; must read yz", alias="type")
        conveyance: Optional[str] = Field(
            "lumped: Lumped, segmented: Vertically segmented. Only the default lumped "
            + "option is allowed if singleValuedZ = no. In the case of lumped conveyance, "
            + "only a single uniform roughness for the whole cross section is allowed, "
            + "i.e., sectionCount must equal 1.",
        )
        yzcount: Optional[str] = Field("Number of YZ-coordinates.", alias="yzCount")
        yCoordinates: Optional[str] = Field(
            "Space separated list of monotonic increasing y-coordinates [m].",
            alias="yCoordinates",
        )
        zCoordinates: Optional[str] = Field(
            "Space separated list of single-valued z-coordinates [m AD].",
            alias="zCoordinates",
        )
        sectioncount: Optional[str] = Field(
            "Number of roughness sections. If the lumped conveyance is selected then "
            + "sectionCount must equal 1.",
            alias="sectionCount",
        )
        frictionpositions: Optional[str] = Field(
            "Locations where the roughness sections start and end. Always one location more than "
            + "sectionCount. The first value should equal 0 and the last value should equal the "
            + "cross section length. Keyword may be skipped if sectionCount = 1.",
            alias="frictionPositions",
        )
        frictionids: Optional[str] = Field(
            "Semicolon separated list of roughness variable names associated with the roughness "
            + "sections. Either this parameter or frictionTypes should be specified. If neither "
            + 'parameter is specified, the frictionIds default to "Main", "FloodPlain1" '
            + 'and "FloodPlain2".',
            alias="frictionIds",
        )
        frictiontypes: Optional[str] = Field(
            "Semicolon separated list of roughness types associated with the roughness sections. "
            + "Either this parameter or frictionIds should be specified. Can be specified as a "
            + "single value if all roughness sections use the same type.",
            alias="frictionTypes",
        )
        frictionvalues: Optional[str] = Field(
            "Space separated list of roughness values; their meaning depends on the roughness "
            + "types selected (only used if frictionTypes specified).",
            alias="frictionValues",
        )

    comments: Comments = Comments()

    type: Literal["yz"] = Field("yz")
    singlevaluedz: Optional[bool] = Field(None, alias="singleValuedZ")
    yzcount: int = Field(alias="yzCount")
    ycoordinates: List[float] = Field(alias="yCoordinates")
    zcoordinates: List[float] = Field(alias="zCoordinates")
    conveyance: Optional[str] = Field("segmented")
    sectioncount: Optional[int] = Field(1, alias="sectionCount")
    frictionpositions: Optional[List[float]] = Field(None, alias="frictionPositions")
    frictionids: Optional[List[str]] = Field(None, alias="frictionIds", delimiter=";")
    frictiontypes: Optional[List[FrictionType]] = Field(
        None, alias="frictionTypes", delimiter=";"
    )
    frictionvalues: Optional[List[float]] = Field(None, alias="frictionValues")

    @field_validator(
        "ycoordinates",
        "zcoordinates",
        "frictionpositions",
        "frictionvalues",
        "frictionids",
        "frictiontypes",
        mode="before",
    )
    @classmethod
    def split_string_list(cls, v, info: ValidationInfo):
        return split_string_on_delimiter(cls, v, info)

    @model_validator(mode="after")
    def check_friction(self):
        self._check_friction_fields(
            self.frictionids, self.frictiontypes, self.frictionvalues, label=self.id
        )
        return self

    @field_validator("frictiontypes", mode="before")
    @classmethod
    def validate_enum_frictiontype(cls, v):
        return enum_value_parser(v, FrictionType)

    @model_validator(mode="after")
    def check_list_lengths_coordinates(self):
        """Validate that the length of the ycoordinates and zcoordinates fields are as expected."""
        validate_correct_length(
            self.model_dump(),
            "ycoordinates",
            "zcoordinates",
            length_name="yzcount",
        )
        return self

    @model_validator(mode="after")
    def check_list_lengths_friction(self):
        """Validate that the length of the frictionids, frictiontypes and frictionvalues field are as expected."""
        validate_correct_length(
            self.model_dump(),
            "frictionids",
            "frictiontypes",
            "frictionvalues",
            length_name="sectioncount",
        )
        return self

    @model_validator(mode="after")
    def check_list_length_frictionpositions(self):
        """Validate that the length of the frictionpositions field is as expected."""
        validate_correct_length(
            self.model_dump(),
            "frictionpositions",
            length_name="sectioncount",
            length_incr=1,  # 1 extra for frictionpositions
        )
        return self


class XYZCrsDef(YZCrsDef, CrossSectionDefinition):
    """XYZCrsDef.

    Crosssection definition with `type=xyz`, to be included in a crossdef file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile.definition[..]`

    All lowercased attributes match with the xyz input as described in
    [UM Sec.C.16.1.5](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.16.1.5).

    This class extends the YZCrsDef class with x-coordinates and an optional
    branchId field. Most other attributes are inherited, but the coordcount
    is overridden under the Pydantic alias "xyzCount".

    Attributes:
        yzcount (Optional[int]): dummy attribute that should not be set nor used.
            Only present to mask the inherited attribute from parent class YZCrsDef.
        xyzcount (int): Number of XYZ-coordinates. Always use this instead of yzcount.
    """

    class Comments(YZCrsDef.Comments):
        """Comments for the XYZCrsDef class."""

        type: Optional[str] = Field("Cross section type; must read xyz", alias="type")
        branchid: Optional[str] = Field(
            "Branch on which the cross section is located.", alias="branchId"
        )
        xyzcount: Optional[str] = Field("Number of XYZ-coordinates.", alias="xyzCount")
        xCoordinates: Optional[str] = Field(
            "Space separated list of x-coordinates [m or degrees East].",
            alias="xCoordinates",
        )
        yCoordinates: Optional[str] = Field(
            "Space separated list of y-coordinates [m or degrees North].",
            alias="yCoordinates",
        )
        zCoordinates: Optional[str] = Field(
            "Space separated list of z-coordinates [m AD].",
            alias="zCoordinates",
        )

    comments: Comments = Comments()

    type: Literal["xyz"] = Field("xyz")
    branchid: Optional[str] = Field(None, alias="branchId")
    yzcount: Optional[int] = Field(
        None, alias="yzCount"
    )  # Trick to not inherit parent's yzcount required field.
    xyzcount: int = Field(alias="xyzCount")
    xcoordinates: List[float] = Field(alias="xCoordinates")

    @field_validator("xcoordinates", mode="before")
    def split_string_xcoordinates(cls, v, info: ValidationInfo):
        return split_string_on_delimiter(cls, v, info)

    @field_validator("xyzcount")
    @classmethod
    def validate_xyzcount_without_yzcount(
        cls, field_value: int, values: ValidationInfo
    ) -> int:
        """Validate the xyzcount field.

        Validates whether this XYZCrsDef does have attribute xyzcount,
        but not the parent class's yzcount.

        Args:
            field_value (Optional[Path]):
                Value given for xyzcount.
            value (int):
                The validated value of xyzcount.

        Raises:
            ValueError:
                When yzcount is present.

        Returns:
            int:
                The value given for xyzcount.
        """
        # Retrieve the algorithm value (if not found use 0).
        yzcount_value = values.data.get("yzcount")
        if field_value is not None and yzcount_value is not None:
            # yzcount should not be set, when xyzcount is set.
            raise ValueError(
                f"xyz cross section definition should not contain field yzCount (rather: xyzCount), current value: {yzcount_value}."
            )
        return field_value

    @model_validator(mode="after")
    def check_list_lengths_coordinates(self):
        """Validate that the length of the xcoordinates, ycoordinates and zcoordinates field are as expected."""
        validate_correct_length(
            self.model_dump(),
            "xcoordinates",
            "ycoordinates",
            "zcoordinates",
            length_name="xyzcount",
        )
        return self


class CrossSection(INIBasedModel):
    """Crosssection.

    A `[CrossSection]` block for use inside a crosssection location file,
    i.e., a [CrossLocModel][hydrolib.core.dflowfm.crosssection.models.CrossLocModel].

    Attributes:
        id (str): Unique cross-section location id.
        branchid (str, optional): Branch on which the cross section is located.
        chainage (str, optional): Chainage on the branch (m).
        x (str, optional): x-coordinate of the location of the cross section.
        y (str, optional): y-coordinate of the location of the cross section.
        shift (float, optional): Vertical shift of the cross section definition [m]. Defined positive upwards.
        definitionid (str): Id of cross section definition.
    """

    class Comments(INIBasedModel.Comments):
        """Comments for the CrossSection class."""

        id: Optional[str] = "Unique cross-section location id."
        branchid: Optional[str] = Field(
            "Branch on which the cross section is located.", alias="branchId"
        )
        chainage: Optional[str] = "Chainage on the branch (m)."

        x: Optional[str] = Field(
            "x-coordinate of the location of the cross section.",
        )
        y: Optional[str] = Field(
            "y-coordinate of the location of the cross section.",
        )
        shift: Optional[str] = Field(
            "Vertical shift of the cross section definition [m]. Defined positive upwards.",
        )
        definitionid: Optional[str] = Field(
            "Id of cross section definition.", alias="definitionId"
        )

    comments: Comments = Comments()
    _header: Literal["CrossSection"] = "CrossSection"
    id: str = Field(alias="id")

    branchid: Optional[str] = Field(None, alias="branchId")
    chainage: Optional[float] = Field(None)

    x: Optional[float] = Field(None)
    y: Optional[float] = Field(None)

    shift: Optional[float] = Field(0.0)
    definitionid: str = Field(alias="definitionId")

    @model_validator(mode="after")
    def validate_that_location_specification_is_correct(self):
        """Validate that the correct location specification is given."""
        validate_location_specification(
            self.model_dump(),
            config=LocationValidationConfiguration(
                validate_node=False,
                validate_num_coordinates=False,
                validate_location_type=False,
            ),
            fields=LocationValidationFieldNames(x_coordinates="x", y_coordinates="y"),
        )
        return self


class CrossLocModel(INIModel):
    """The overall crosssection location model that contains the contents of one crossloc file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crosslocfile`.

    Attributes:
        general (CrossLocGeneral):
            `[General]` block with file metadata.
        crosssection (List[CrossSection]):
            List of `[CrossSection]` blocks for all cross-section locations, The crosssection attribute also accepts
            single cross section.

    Examples:
        - Create the read `CrossLocModel` class from file.
        ```python
        >>> from hydrolib.core.dflowfm.crosssection.models import CrossLocModel
        >>> from pathlib import Path
        >>> path = Path("examples/data/crsloc.ini")
        >>> crossloc_model = CrossLocModel(path)
        >>> print(len(crossloc_model.crosssection))
        2
        >>> print(crossloc_model.crosssection[0])
        comments=Comments(id=None, branchid=None, chainage=None, x='x-coordinate of the location of the cross section.', y='y-coordinate of the location of the cross section.', shift=None, definitionid=None) id='Channel1_50.000' branchid='Channel1' chainage=50.0 x=None y=None shift=1.0 definitionid='Prof1'

        ```

        - Create the `CrossLocModel` class by providing values for the `crosssection` attribute.
        ```python
        >>> data = {
        ...    "id": 99,
        ...    "branchId": 9,
        ...    "chainage": 403,
        ...    "shift": 0.0,
        ...    "definitionId": 99
        ... }
        >>> cross_section = CrossSection(**data)
        >>> crossloc = CrossLocModel(crosssection=cross_section)
        >>> type(crossloc.crosssection)
        <class 'list'>
        >>> len(crossloc.crosssection)
        1

        ```

        - Create the `CrossLocModel` class by providing values as a dictionary.
        ```python
        >>> data = {
        ...     "crosssection": {
        ...         "id": 99,
        ...         "branchId": 9,
        ...         "chainage": 403.089709,
        ...         "shift": 0.0,
        ...         "definitionId": 99,
        ...     }
        ... }
        >>> crossloc = CrossLocModel(**data)
        >>> print(crossloc.crosssection)
        [CrossSection(comments=Comments(id='Unique cross-section location id.', branchid='Branch on which the cross section is located.', chainage='Chainage on the branch (m).', x='x-coordinate of the location of the cross section.', y='y-coordinate of the location of the cross section.', shift='Vertical shift of the cross section definition [m]. Defined positive upwards.', definitionid='Id of cross section definition.'), id='99', branchid='9', chainage=403.089709, x=None, y=None, shift=0.0, definitionid='99')]

        ```
    """

    general: CrossLocGeneral = CrossLocGeneral()
    crosssection: List[CrossSection] = Field(default_factory=list)

    @classmethod
    def _filename(cls) -> str:
        return "crsloc"

    @field_validator("crosssection", mode="before")
    def ensure_crosssection_is_list(cls, v):
        """Converting the crosssection to a list if it is not already a list."""
        if isinstance(v, list):
            return v
        elif v is None:
            return []
        return [v]


CrossSectionDefinitionUnion = Annotated[
    Union[
        CircleCrsDef,
        RectangleCrsDef,
        ZWRiverCrsDef,
        ZWCrsDef,
        YZCrsDef,
        XYZCrsDef,
    ],
    Field(discriminator="type"),
]


class CrossDefModel(INIModel):
    """
    The overall crosssection definition model that contains the contents of one crossdef file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.crossdeffile`.

    Attributes:
        general (CrossdefGeneral): `[General]` block with file metadata.
        definition (List[CrossSectionDefinition]): List of `[Definition]` blocks for all cross sections.
    """

    general: CrossDefGeneral = CrossDefGeneral()
    definition: Annotated[
        List[CrossSectionDefinitionUnion], BeforeValidator(make_list)
    ] = []

    @classmethod
    def _filename(cls) -> str:
        return "crsdef"



"""util.py provides additional utility methods related to handling ini files.
"""

from datetime import datetime
from enum import Enum
from operator import eq
from typing import Any, Callable, Dict, List, Optional, Set, Type

from pydantic import BaseModel, ValidationInfo
from pydantic.fields import FieldInfo

from hydrolib.core.base.utils import operator_str, str_is_empty_or_none, to_list
from hydrolib.core.dflowfm.common.models import LocationType


def split_string_on_delimiter(cls, v: Any, field: ValidationInfo):
    """Split a string on the list field delimiter, and return a list of strings.

    If the input is a string, it is split on the delimiter defined in the class.
    If the input is anything else, it is returned as is.

    Args:
        cls (Type[BaseModel]): The class that contains the field.
        v (Any): The value to split.
        field (ValidationInfo): The field information.

    Returns:
        List[str] or Any: A list of strings if the input was a string, otherwise
            the input value as is.
    """
    if isinstance(v, str):
        v = v.split(cls.get_list_field_delimiter(field.field_name))
        v = [item.strip() for item in v if item != ""]
    return v


def enum_value_parser(
    v,
    enum: Type[Enum],
    alternative_enum_values: Optional[Dict[str, List[str]]] = None,
):
    """Return a function that converts strings (and string lists) to Enum values.

    Args:
        enum (Type[Enum]): The Enum type to parse values into.
        alternative_enum_values (Optional[Dict[str, List[str]]]): A dictionary mapping enum values
            to alternative string representations. If provided, the parser will also accept these
            alternative strings as valid inputs for the corresponding enum values.

    Returns:
        Callable: A function that takes a value (or list of values) and returns the corresponding
            Enum value or raises a ValueError if the value is invalid.

    Raises:
        ValueError: If the input value is not a valid Enum value or does not match any
            alternative string representations.
    """
    if isinstance(v, list):
        result = [parse_enum(item, enum, alternative_enum_values) for item in v]
    else:
        result = parse_enum(v, enum, alternative_enum_values)
    return result


def parse_enum(
    v, enum: Type[Enum], alternative_enum_values: Optional[Dict[str, List[str]]] = None
):
    result = None

    if isinstance(v, enum):
        result = v
    elif isinstance(v, str):
        v_lower = v.lower()
        for entry in enum:
            if entry.value.lower() == v_lower:
                result = entry
                break
            if (
                alternative_enum_values
                and (alts := alternative_enum_values.get(entry.value))
                and any(v_lower == alt.lower() for alt in alts)
            ):
                result = entry
                break
    if result is None:
        valid_values = [e.value for e in enum]
        raise ValueError(f"Invalid enum value: {v!r}. Expected one of: {valid_values}")
    return result


def ensure_list(v: Any):
    """Ensure that the input is a list.

    Args:
        v (Any): The value to ensure is a list.

    Returns:
        List[Any]: A list containing the input value if it was a dictionary,
            or the input value itself if it was already a list.

    Raises:
        TypeError: If the input is not a list or a dictionary.
    """
    if isinstance(v, dict):
        v = [v]
    if not isinstance(v, list):
        raise TypeError("Expected a list or a single dictionary")
    return v


def make_list(v: Any):
    if not isinstance(v, list):
        v = [v]
    return v


def validate_correct_length(
    values: Dict,
    *field_names,
    length_name: str,
    length_incr: int = 0,
    list_required_with_length: bool = False,
    min_length: int = 0,
):
    """
    Validate the correct length (and presence) of several list fields in an object.

    Args:
        values (Dict):
            dictionary of values to validate.
        *field_names (str):
            names of the instance variables that are a list and need checking.
        length_name (str):
            name of the instance variable that stores the expected length.
        length_incr (int):
            Optional extra increment of length value (e.g., to have +1 extra value in lists).
        list_required_with_length (obj:`bool`, optional):
            Whether each list *must* be present if the length attribute is present (and > 0) in the input values.
            Default: False. If False, list length is only checked for the lists that are not None.
        min_length (int):
            minimum for list length value, overrides length_name value if that is smaller. For example, to require
            list length 1 when length value is given as 0.

    Raises:
        ValueError:
            When the number of values for any of the given field_names is not as expected.

    Returns:
        Dict:
            Dictionary of validated values.
    """

    def _get_incorrect_length_validation_message() -> str:
        """Make a string with a validation message, ready to be format()ed with
        field name and length name."""
        incrstring = f" + {length_incr}" if length_incr != 0 else ""
        minstring = f" (and at least {min_length})" if min_length > 0 else ""

        return (
            "Number of values for {} should be equal to the {} value"
            + incrstring
            + minstring
            + "."
        )

    def _validate_listfield_length(
        field_name: str, field: Optional[List[Any]], requiredlength: int
    ):
        """Validate the length of a single field, which should be a list."""

        if field is not None and len(field) != requiredlength:
            raise ValueError(
                _get_incorrect_length_validation_message().format(
                    field_name, length_name
                )
            )
        if field is None and list_required_with_length and requiredlength > 0:
            raise ValueError(
                f"List {field_name} cannot be missing if {length_name} is given."
            )

    length = values.get(length_name)
    if length is None:
        # length attribute not present, possibly defer validation to a subclass.
        return values

    requiredlength = max(int(length) + length_incr, min_length)

    for field_name in field_names:
        field = values.get(field_name)
        _validate_listfield_length(field_name, field, requiredlength)


def validate_forbidden_fields(
    values: Dict,
    *field_names,
    conditional_field_name: str,
    conditional_value: Any,
    comparison_func: Callable[[Any, Any], bool] = eq,
) -> Dict:
    """
    Validates whether certain fields are *not* provided, if `conditional_field_name` is equal to `conditional_value`.
    The equality check can be overridden with another comparison operator function.

    Args:
        values (Dict): Dictionary of input class fields.
        *field_names (str): Names of the instance variables that need to be validated.
        conditional_field_name (str): Name of the instance variable on which the fields are dependent.
        conditional_value (Any): Value that the conditional field should contain to perform this validation.
        comparison_func (Callable): binary operator function, used to override the default "eq" check for the conditional field value.

    Raises:
        ValueError: When a forbidden field is provided.

    Returns:
        Dict: Validated dictionary of input class fields.
    """
    if (val := values.get(conditional_field_name)) is None or not comparison_func(
        val, conditional_value
    ):
        return values

    for field in field_names:
        if values.get(field) is not None:
            raise ValueError(
                f"{field} is forbidden when {conditional_field_name} {operator_str(comparison_func)} {conditional_value}"
            )

    return values


def validate_required_fields(
    values: Dict,
    *field_names,
    conditional_field_name: str,
    conditional_value: Any,
    comparison_func: Callable[[Any, Any], bool] = eq,
):
    """
    Validates whether the specified fields are provided, if `conditional_field_name` is equal to `conditional_value`.
    The equality check can be overridden with another comparison operator function.

    Args:
        values (Dict): Dictionary of input class fields.
        *field_names (str): Names of the instance variables that need to be validated.
        conditional_field_name (str): Name of the instance variable on which the fields are dependent.
        conditional_value (Any): Value that the conditional field should contain to perform this validation.
        comparison_func (Callable): binary operator function, used to override the default "eq" check for the conditional field value.

    Raises:
        ValueError: When a required field is not provided under the given conditions.

    Returns:
        Dict: Validated dictionary of input class fields.
    """

    if (val := values.get(conditional_field_name)) is None or not comparison_func(
        val, conditional_value
    ):
        return values

    for field in field_names:
        if values.get(field) is None:
            raise ValueError(
                f"{field} should be provided when {conditional_field_name} {operator_str(comparison_func)} {conditional_value}"
            )


def validate_conditionally(
    values: Dict,
    root_vldt: classmethod,
    conditional_field_name: str,
    conditional_value: Any,
    comparison_func: Callable[[Any, Any], bool] = eq,
) -> Dict:
    """
    Validate whether certain fields are *not* provided, if `conditional_field_name` is equal to `conditional_value`.
    The equality check can be overridden with another comparison operator function.

    Args:
        values (Dict):
            Dictionary of input class fields.
        root_vldt (classmethod):
            A root validator that is to be called *if* the condition is satisfied.
        conditional_field_name (str):
            Name of the instance variable that determines whether the root validator must be called or not.
        conditional_value (Any):
            Value that the conditional field should be compared with to perform this validation.
        comparison_func (Callable):
            Binary operator function, used to override the default "eq" check for the conditional field value.

    Returns:
        Dict:
            Validated dictionary of input class fields.
    """
    if (val := values.get(conditional_field_name)) is not None and comparison_func(
        val, conditional_value
    ):
        # Condition is met: call the actual root validator, passing on the attribute values.
        root_vldt(values)


def validate_datetime_string(
    field_value: Optional[str], field: ValidationInfo
) -> Optional[str]:
    """Validate that a field value matches the expected datetime format.

    The validation checks that the date formats conform to either 'YYYYmmddHHMMSS' or 'YYYYmmdd'.

    Args:
        field_value (Optional[str]): value of a Pydantic field, may be optional.
        field (ModelField): the underlying Pydantic ModelField, used in error
            message.

    Returns:
        Optional[str]: the original input value, if valid.

    Raises:
        ValueError: if a non-empty input string does not have valid format.
    """
    if (
        field_value is not None
        and len(field_value.strip()) > 0
        and field_value != "yyyymmddhhmmss"
    ):
        formats = {14: "%Y%m%d%H%M%S", 8: "%Y%m%d"}
        result = False
        format_length = len(field_value)
        if format_length in formats:
            try:
                datetime.strptime(field_value, formats[format_length])
                result = True
            except ValueError:
                pass

        if not result:
            raise ValueError(
                f"Invalid datetime string for {field.field_name}: '{field_value}', expecting 'YYYYmmddHHMMSS' or 'YYYYmmdd'."
            )

    return field_value  # this is the value written to the class field


def get_from_subclass_defaults(cls: Type[BaseModel], fieldname: str, value: str) -> str:
    """
    Gets a value that corresponds with the default field value of one of the subclasses.
    If the subclass doesn't have the specified field, it will look into its own subclasses
    recursively for the specified fieldname.

    Args:
        cls (Type[BaseModel]): The parent model type.
        fieldname (str): The field name for which retrieve the default for.
        value (str): The value to compare with.

    Returns:
        str: The field default that corresponds to the value. If nothing is found return the input value.
    """
    # Immediately check in direct subclasses, not in base cls itself:
    for c in cls.__subclasses__():
        default = _try_get_default_value(c, fieldname, value)
        if default is not None:
            return default

    # No matching default was found, return input value:
    return value


def _try_get_default_value(
    c: Type[BaseModel], fieldname: str, value: str
) -> Optional[str]:
    """Helper subroutine to get the default value for a particular field in
    the given class or any of its descendant classes, if it matches the input
    value (case insensitive).

    This method recurses depth-first topdown into the class'es subclasses.

        c (Type[BaseModel]): The base model type where the search starts.
        fieldname (str): The field name for which retrieve the default for.
        value (str): The value to compare with.

    Returns:
        Optional[str]: The field default that corresponds to the value. If nothing is found return None.
    """
    stack = [c]
    result = None
    while stack:
        current_class = stack.pop()
        if not hasattr(current_class, "model_fields"):
            continue
        field = current_class.model_fields.get(fieldname)
        if field is not None:
            # In pydantic v2, default is accessed through default_factory or directly
            if hasattr(field, "default_factory") and field.default_factory is not None:
                default = field.default_factory()
            else:
                default = field.default

            if (
                default is not None
                and hasattr(default, "lower")
                and default.lower() == value.lower()
            ):
                result = default
                break
        # Add subclasses to stack for further checking
        stack.extend(current_class.__subclasses__())
    return result


class LocationValidationConfiguration(BaseModel):
    """Class that holds the various configuration settings needed for location validation."""

    validate_node: bool = True
    """bool, optional: Whether or not node location specification should be validated. Defaults to True."""

    validate_coordinates: bool = True
    """bool, optional: Whether or not coordinate location specification should be validated. Defaults to True."""

    validate_branch: bool = True
    """bool, optional: Whether or not branch location specification should be validated. Defaults to True."""

    validate_num_coordinates: bool = True
    """bool, optional: Whether or not the number of coordinates should be validated or not. This option is only relevant when `validate_coordinates` is True. Defaults to True."""

    validate_location_type: bool = True
    """bool, optional: Whether or not the location type should be validated. Defaults to True."""

    minimum_num_coordinates: int = 0
    """int, optional: The minimum required number of coordinates. This option is only relevant when `validate_coordinates` is True. Defaults to 0."""


class LocationValidationFieldNames(BaseModel):
    """Class that holds the various field names needed for location validation."""

    node_id: str = "nodeId"
    """str, optional: The node id field name. Defaults to `nodeId`."""

    branch_id: str = "branchId"
    """str, optional: The branch id field name. Defaults to `branchId`."""

    chainage: str = "chainage"
    """str, optional: The chainage field name. Defaults to `chainage`."""

    x_coordinates: str = "xCoordinates"
    """str, optional: The x-coordinates field name. Defaults to `xCoordinates`."""

    y_coordinates: str = "yCoordinates"
    """str, optional: The y-coordinates field name. Defaults to `yCoordinates`."""

    num_coordinates: str = "numCoordinates"
    """str, optional: The number of coordinates field name. Defaults to `numCoordinates`."""

    location_type: str = "locationType"
    """str, optional: The location type field name. Defaults to `locationType`."""


def validate_location_specification(
    values: Dict,
    config: Optional[LocationValidationConfiguration] = None,
    fields: Optional[LocationValidationFieldNames] = None,
) -> Dict:
    """
    Validates whether the correct location specification is given in
    typical 1D2D input in an IniBasedModel class.

    Validates for presence of at least one of: nodeId, branchId with chainage,
    xCoordinates with yCoordinates, or xCoordinates with yCoordinates and numCoordinates.
    Validates for the locationType for nodeId and branchId.

    Args:
        values (Dict):
            Dictionary of object's validated fields.
        config (LocationValidationConfiguration, optional):
            Configuration for the location validation. Default is None.
        fields (LocationValidationFieldNames, optional):
            Fields names that should be used for the location validation. Default is None.

    Raises:
        ValueError: When exactly one of the following combinations were not given:
            - nodeId
            - branchId with chainage
            - xCoordinates with yCoordinates
            - xCoordinates with yCoordinates and numCoordinates.
        ValueError: When numCoordinates does not meet the requirement minimum amount or does not match the amount of xCoordinates or yCoordinates.
        ValueError: When locationType should be 1d but other was specified.

    Returns:
        Dict: Validated dictionary of input class fields.
    """

    if config is None:
        config = LocationValidationConfiguration()

    if fields is None:
        fields = LocationValidationFieldNames()

    has_node_id = not str_is_empty_or_none(values.get(fields.node_id.lower()))
    has_branch_id = not str_is_empty_or_none(values.get(fields.branch_id.lower()))
    has_chainage = values.get(fields.chainage.lower()) is not None
    has_x_coordinates = values.get(fields.x_coordinates.lower()) is not None
    has_y_coordinates = values.get(fields.y_coordinates.lower()) is not None
    has_num_coordinates = values.get(fields.num_coordinates.lower()) is not None

    # ----- Local validation functions
    def get_length(field: str):
        value = values[field.lower()]
        return len(to_list(value))

    def validate_location_type(expected_location_type: LocationType) -> None:
        location_type = values.get(fields.location_type.lower(), None)
        if str_is_empty_or_none(location_type):
            values[fields.location_type.lower()] = expected_location_type
        elif location_type != expected_location_type:
            raise ValueError(
                f"{fields.location_type} should be {expected_location_type} but was {location_type}"
            )

    def validate_coordinates_with_num_coordinates() -> None:
        length_x_coordinates = get_length(fields.x_coordinates)
        length_y_coordinates = get_length(fields.y_coordinates)
        num_coordinates = values[fields.num_coordinates.lower()]

        if not num_coordinates == length_x_coordinates == length_y_coordinates:
            raise ValueError(
                f"{fields.num_coordinates} should be equal to the amount of {fields.x_coordinates} and {fields.y_coordinates}"
            )

        validate_minimum_num_coordinates(num_coordinates)

    def validate_coordinates() -> None:
        len_x_coordinates = get_length(fields.x_coordinates)
        len_y_coordinates = get_length(fields.y_coordinates)

        if len_x_coordinates != len_y_coordinates:
            raise ValueError(
                f"{fields.x_coordinates} and {fields.y_coordinates} should have an equal amount of coordinates"
            )

        validate_minimum_num_coordinates(len_x_coordinates)

    def validate_minimum_num_coordinates(actual_num: int) -> None:
        if actual_num < config.minimum_num_coordinates:
            raise ValueError(
                f"{fields.x_coordinates} and {fields.y_coordinates} should have at least {config.minimum_num_coordinates} coordinate(s)"
            )

    def is_valid_node_specification() -> bool:
        has_other = (
            has_branch_id
            or has_chainage
            or has_x_coordinates
            or has_y_coordinates
            or has_num_coordinates
        )
        return has_node_id and not has_other

    def is_valid_branch_specification() -> bool:
        has_other = (
            has_node_id or has_x_coordinates or has_y_coordinates or has_num_coordinates
        )
        return has_branch_id and has_chainage and not has_other

    def is_valid_coordinates_specification() -> bool:
        has_other = has_node_id or has_branch_id or has_chainage or has_num_coordinates
        return has_x_coordinates and has_y_coordinates and not has_other

    def is_valid_coordinates_with_num_coordinates_specification() -> bool:
        has_other = has_node_id or has_branch_id or has_chainage
        return (
            has_x_coordinates
            and has_y_coordinates
            and has_num_coordinates
            and not has_other
        )

    # -----

    error_parts: List[str] = []

    if config.validate_node:
        if is_valid_node_specification():
            if config.validate_location_type:
                validate_location_type(LocationType.oned)
            return values

        error_parts.append(fields.node_id)

    if config.validate_branch:
        if is_valid_branch_specification():
            if config.validate_location_type:
                validate_location_type(LocationType.oned)
            return values

        error_parts.append(f"{fields.branch_id} and {fields.chainage}")

    if config.validate_coordinates:
        if config.validate_num_coordinates:
            if is_valid_coordinates_with_num_coordinates_specification():
                validate_coordinates_with_num_coordinates()
                return values

            error_parts.append(
                f"{fields.x_coordinates}, {fields.y_coordinates} and {fields.num_coordinates}"
            )

        else:
            if is_valid_coordinates_specification():
                validate_coordinates()
                return values

            error_parts.append(f"{fields.x_coordinates} and {fields.y_coordinates}")

    error = " or ".join(error_parts) + " should be provided"
    raise ValueError(error)


def rename_keys_for_backwards_compatibility(
    values: Dict, keys_to_rename: Dict[str, List[str]]
) -> Dict:
    """
    Renames the provided keys to support backwards compatibility.

    Args:

        values (Dict): Dictionary of input class fields.
        keys_to_rename (Dict[str, List[str]]): Dictionary of keys and a list of old keys that
        should be converted to the current key.

    Returns:
        Dict: Dictionary where the provided keys are renamed.
    """
    if not isinstance(values, dict):
        return values
    for current_keyword, old_keywords in keys_to_rename.items():
        if current_keyword in values:
            continue

        for old_keyword in old_keywords:
            if (value := values.get(old_keyword)) is not None:
                values[current_keyword] = value
                del values[old_keyword]
                break

    return values


class UnknownKeywordErrorManager:
    """
    Error manager for unknown keys.
    Detects unknown keys and manages the Error to the user.
    """

    def raise_error_for_unknown_keywords(
        self,
        data: Dict[str, Any],
        section_header: str,
        fields: Dict[str, FieldInfo],
        excluded_fields: Set[str],
    ) -> None:
        """
        Notify the user of unknown keywords.

        Args:
            data (Dict[str, Any]):
                Input data containing all properties which are checked on unknown keywords.
            section_header (str):
                Header of the section in which unknown keys might be detected.
            fields (Dict[str, FieldInfo]):
                Known fields of the section.
            excluded_fields (Set[str]):
                Fields which should be excluded from the check for unknown keywords.
        """
        unknown_keywords = self._get_all_unknown_keywords(data, fields, excluded_fields)

        if len(unknown_keywords) > 0:
            raise ValueError(
                f"Unknown keywords are detected in section: '{section_header}', '{unknown_keywords}'"
            )

    def _get_all_unknown_keywords(
        self,
        data: Dict[str, Any],
        fields: Dict[str, FieldInfo],
        excluded_fields: Set,
    ) -> List[str]:
        """
        Get all unknown keywords in the data.

        Args:
            data: Dict[str, Any]:
                Input data containing all properties which are checked on unknown keywords.
            fields: Dict[str, FieldInfo]:
                Known fields of the Model.
            excluded_fields: Set[str]:
                Fields which should be excluded from the check for unknown keywords.

        Returns:
            List[str]: List of unknown keywords.
        """
        list_of_unknown_keywords = []
        for keyword in data:
            if self._is_unknown_keyword(keyword, fields, excluded_fields):
                list_of_unknown_keywords.append(keyword)

        return list_of_unknown_keywords

    @staticmethod
    def _is_unknown_keyword(
        keyword: str, fields: Dict[str, FieldInfo], excluded_fields: Set
    ) -> bool:
        """
        Check if the given field name equals to any of the model field names or aliases, if not, the function checks if
        the field is not in the excluded_fields parameter.

        Args:
            keyword: str: Name of the field.
            fields: Dict[str, FieldInfo]: Known fields of the Model.
            excluded_fields: Set[str]: Fields which should be excluded from the check for unknown keywords.

        Returns:
            bool: True if the field is unknown (not a field name or alias and and not in the exclude list),
            False otherwise
        """
        exists = keyword in fields or any(
            hasattr(field_info, "alias") and keyword == field_info.alias
            for field_info in fields.values()
        )
        # the field is not in the known fields, check if it should be excluded
        unknown = not exists and keyword not in excluded_fields

        return unknown



from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

from pydantic import Field, field_validator

from hydrolib.core.base.models import BaseModel
from hydrolib.core.base.utils import to_key


class CommentBlock(BaseModel):
    """CommentBlock defines a comment block within a deltares ini file.

    Attributes:
        lines (List[str]): The actual lines of the CommentBlock
    """

    lines: List[str] = Field(default_factory=list)


class Property(BaseModel):
    """Property defines a deltares ini property

    Attributes:
        key (str): The key of this Property
        value (Optional[str]): The value associated with this Property
        comment (Optional[str]): The comment associated with this Property
    """

    key: str
    value: Optional[str] = None
    comment: Optional[str] = None

    def get_item(self):
        return {self.key: self.value}

    def get_comment(self):
        return {self.key: self.comment}

    @field_validator("value", mode="before")
    def make_value_str(cls, value: Any) -> Optional[str]:
        """Ensure that the value is always a string, even if it is None."""
        if value is None:
            return None
        return f"{value}"


ContentElement = Union[Property, CommentBlock]
DatablockRow = Sequence[str]
Datablock = Sequence[DatablockRow]


class Section(BaseModel):
    """Section defines a deltares ini section

    A deltares ini section consists of a header and an ordered list of
    properties and comments.

    For example:

    ```
    [header-title]
    a=1
    b=2
    # and a comment
    c=4
    ```

    An optional datablock can be defined (used in for example .bc files).

    Attributes:
        header (str): The header (without brackets) of this Section
        content (List[ContentElement]):
            The ordered list of Property and CommentBlock objects
        datablock (Optional[Datablock]):
            An optional data block associated with this Section. The datablock is
            structured as a sequence of rows, i.e. datablock[2][1] refers to the third
            row, second column.

    Examples:
        ```python
        >>> from hydrolib.core.dflowfm.ini.io_models import Section
        >>> section = Section(header="header", content=[Property(key="a", value="1")])
        >>> section.flatten()
        {'_header': 'header', 'datablock': None, 'a': '1', 'comments': {'a': None}}

        ```
    """

    header: str = Field(alias="_header")
    content: List[ContentElement] = Field(default_factory=list)

    # these are primarily relevant for bc files
    datablock: Optional[Datablock] = None

    def model_dump(self, *args, **kwargs):
        kwargs["by_alias"] = True
        return super().model_dump(*args, **kwargs)

    def flatten(self, duplicate_key_as_list=True, with_comments=True) -> Dict:
        converted_content = self._convert_section_content(
            duplicate_key_as_list, with_comments
        )
        underlying_dict = self._convert_section_to_dict()
        return {**underlying_dict, **converted_content}

    def _convert_section_to_dict(self) -> Dict:
        return self.model_dump(
            exclude={
                "start_line",
                "end_line",
                "content",
            }
        )

    def _convert_section_content(
        self, duplicate_key_as_list: bool, with_comments: bool
    ) -> Dict:
        def group_and_flatten(val: Iterable[Tuple[str, Any]]) -> Dict[str, Any]:
            if duplicate_key_as_list:
                return reduce(_combine_in_dict, val, {})
            else:
                return dict(val)

        values = group_and_flatten(
            (to_key(v.key), v.value) for v in self.content if isinstance(v, Property)
        )

        if with_comments:
            values["comments"] = group_and_flatten(
                (to_key(v.key), v.comment)
                for v in self.content
                if isinstance(v, Property)
            )

        return values


class Document(BaseModel):
    """Document defines a Deltares ini document

    Attributes:
        header_comment (List[CommentBlock]):
            An ordered list of comment blocks defined in the header of the Document
        sections (List[Section]):
            An ordered list of sections defined in this Document
    """

    header_comment: List[CommentBlock] = []
    sections: List[Section] = []

    def flatten(self, duplicate_key_as_list=True, with_comments=True):
        data = {}
        for v in self.sections:
            key = to_key(v.header)
            value = v.flatten(duplicate_key_as_list, with_comments)
            if key in data and not isinstance(data[key], list):
                data[key] = [data[key], value]
            elif key in data:
                data[key].append(value)
            else:
                data[key] = value
        return data


def _combine_in_dict(
    dictionary: Dict[str, Any], key_value_pair: Tuple[str, Any]
) -> Dict[str, Any]:
    key, value = key_value_pair

    if key in dictionary:
        if not isinstance(dictionary[key], List):
            dictionary[key] = [
                dictionary[key],
            ]
        dictionary[key].append(value)
    else:
        dictionary[key] = value

    return dictionary
